import re
from itertools import chain
from typing import (
    Optional,
    Union,
    Sequence,
    Tuple,
    Any,
    Container,
    List,
    Iterable,
    Iterator,
    Callable,
    cast,
    TYPE_CHECKING,
)

from debputy.linting.lint_util import LintState, te_position_to_lsp
from debputy.lsp.debputy_ls import DebputyLanguageServer
from debputy.lsp.lsp_debian_control_reference_data import (
    Deb822FileMetadata,
    Deb822KnownField,
    StanzaMetadata,
    F,
    S,
    SUBSTVAR_RE,
)
from debputy.lsp.lsp_features import SEMANTIC_TOKEN_TYPES_IDS
from debputy.lsp.quickfixes import propose_correct_text_quick_fix
from debputy.lsp.text_util import (
    trim_end_of_line_whitespace,
    SemanticTokensState,
)
from debputy.lsp.vendoring._deb822_repro.locatable import (
    START_POSITION,
    Range as TERange,
    Position as TEPosition,
)
from debputy.lsp.vendoring._deb822_repro.parsing import (
    Deb822KeyValuePairElement,
    Deb822ParagraphElement,
    Deb822FileElement,
    Deb822CommentElement,
)
from debputy.lsp.vendoring._deb822_repro.tokens import tokenize_deb822_file, Deb822Token
from debputy.lsp.vendoring._deb822_repro.types import TokenOrElement
from debputy.lsprotocol.types import (
    CompletionParams,
    CompletionList,
    CompletionItem,
    Position,
    MarkupContent,
    Hover,
    MarkupKind,
    HoverParams,
    FoldingRangeParams,
    FoldingRange,
    FoldingRangeKind,
    SemanticTokensParams,
    SemanticTokens,
    TextEdit,
    MessageType,
    SemanticTokenTypes,
)
from debputy.util import _info, _warn

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types


try:
    from pygls.server import LanguageServer
    from pygls.workspace import TextDocument
except ImportError:
    pass


_CONTAINS_SPACE_OR_COLON = re.compile(r"[\s:]")


def in_range(
    te_range: TERange,
    cursor_position: Position,
    *,
    inclusive_end: bool = False,
) -> bool:
    cursor_line = cursor_position.line
    start_pos = te_range.start_pos
    end_pos = te_range.end_pos
    if cursor_line < start_pos.line_position or cursor_line > end_pos.line_position:
        return False

    if start_pos.line_position == end_pos.line_position:
        start_col = start_pos.cursor_position
        cursor_col = cursor_position.character
        end_col = end_pos.cursor_position
        if inclusive_end:
            return start_col <= cursor_col <= end_col
        return start_col <= cursor_col < end_col

    if cursor_line == end_pos.line_position:
        return cursor_position.character < end_pos.cursor_position

    return (
        cursor_line > start_pos.line_position
        or start_pos.cursor_position <= cursor_position.character
    )


def _field_at_position(
    stanza: Deb822ParagraphElement,
    stanza_metadata: S,
    stanza_range: TERange,
    position: Position,
) -> Tuple[Optional[Deb822KeyValuePairElement], Optional[F], str, bool]:
    te_range = TERange(stanza_range.start_pos, stanza_range.start_pos)
    for token_or_element in stanza.iter_parts():
        te_range = token_or_element.size().relative_to(te_range.end_pos)
        if not in_range(te_range, position):
            continue
        if isinstance(token_or_element, Deb822KeyValuePairElement):
            value_range = token_or_element.value_element.range_in_parent().relative_to(
                te_range.start_pos
            )
            known_field = stanza_metadata.get(token_or_element.field_name)
            in_value = in_range(value_range, position)
            interpreter = (
                known_field.field_value_class.interpreter()
                if known_field is not None
                else None
            )
            matched_value = ""
            if in_value and interpreter is not None:
                interpreted = token_or_element.interpret_as(interpreter)
                for value_ref in interpreted.iter_value_references():
                    value_token_range = (
                        value_ref.locatable.range_in_parent().relative_to(
                            value_range.start_pos
                        )
                    )
                    if in_range(value_token_range, position, inclusive_end=True):
                        matched_value = value_ref.value
                        break
            return token_or_element, known_field, matched_value, in_value
    return None, None, "", False


def _allow_stanza_continuation(
    token_or_element: TokenOrElement,
    is_completion: bool,
) -> bool:
    if not is_completion:
        return False
    if token_or_element.is_error or token_or_element.is_comment:
        return True
    return (
        token_or_element.is_whitespace
        and token_or_element.convert_to_text().count("\n") < 2
    )


def _at_cursor(
    deb822_file: Deb822FileElement,
    file_metadata: Deb822FileMetadata[S, F],
    doc: "TextDocument",
    lines: List[str],
    client_position: Position,
    is_completion: bool = False,
) -> Tuple[
    Position,
    Optional[str],
    str,
    bool,
    Optional[S],
    Optional[F],
    Iterable[Deb822ParagraphElement],
]:
    server_position = doc.position_codec.position_from_client_units(
        lines,
        client_position,
    )
    te_range = TERange(
        START_POSITION,
        START_POSITION,
    )
    paragraph_no = -1
    previous_stanza: Optional[Deb822ParagraphElement] = None
    next_stanza: Optional[Deb822ParagraphElement] = None
    current_word = doc.word_at_position(client_position)
    in_value: bool = False
    file_iter = iter(deb822_file.iter_parts())
    matched_token: Optional[TokenOrElement] = None
    matched_field: Optional[str] = None
    stanza_metadata: Optional[S] = None
    known_field: Optional[F] = None

    for token_or_element in file_iter:
        te_range = token_or_element.size().relative_to(te_range.end_pos)
        if isinstance(token_or_element, Deb822ParagraphElement):
            previous_stanza = token_or_element
            paragraph_no += 1
        elif not _allow_stanza_continuation(token_or_element, is_completion):
            previous_stanza = None
        if not in_range(te_range, server_position):
            continue
        matched_token = token_or_element
        if isinstance(token_or_element, Deb822ParagraphElement):
            stanza_metadata = file_metadata.guess_stanza_classification_by_idx(
                paragraph_no
            )
            kvpair, known_field, current_word, in_value = _field_at_position(
                token_or_element,
                stanza_metadata,
                te_range,
                server_position,
            )
            if kvpair is not None:
                matched_field = kvpair.field_name
        break

    if matched_token is not None and _allow_stanza_continuation(
        matched_token,
        is_completion,
    ):
        next_te = next(file_iter, None)
        if isinstance(next_te, Deb822ParagraphElement):
            next_stanza = next_te

    stanza_parts = (p for p in (previous_stanza, next_stanza) if p is not None)

    if stanza_metadata is None and is_completion:
        if paragraph_no < 0:
            paragraph_no = 0
        stanza_metadata = file_metadata.guess_stanza_classification_by_idx(paragraph_no)

    return (
        server_position,
        matched_field,
        current_word,
        in_value,
        stanza_metadata,
        known_field,
        stanza_parts,
    )


def deb822_completer(
    ls: "DebputyLanguageServer",
    params: CompletionParams,
    file_metadata: Deb822FileMetadata[Any, Any],
) -> Optional[Union[CompletionList, Sequence[CompletionItem]]]:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lines = doc.lines
    lint_state = ls.lint_state(doc)
    deb822_file = lint_state.parsed_deb822_file_content
    if deb822_file is None:
        _warn("The deb822 result missing failed!?")
        ls.show_message_log(
            "Internal error; could not get deb822 content!?", MessageType.Warning
        )
        return None

    (
        server_pos,
        current_field,
        word_at_position,
        in_value,
        stanza_metadata,
        known_field,
        matched_stanzas,
    ) = _at_cursor(
        deb822_file,
        file_metadata,
        doc,
        lines,
        params.position,
        is_completion=True,
    )

    if lines[server_pos.line].startswith("#"):
        return

    items: Optional[Sequence[CompletionItem]]
    markdown_kind = ls.completion_item_document_markup(
        MarkupKind.Markdown, MarkupKind.PlainText
    )
    if in_value:
        _info(f"Completion for field value {current_field} -- {word_at_position}")
        if known_field is None:
            return None
        value_being_completed = word_at_position
        items = known_field.value_options_for_completer(
            lint_state,
            list(matched_stanzas),
            value_being_completed,
            markdown_kind,
        )
    else:
        _info("Completing field name")
        assert stanza_metadata is not None
        items = _complete_field_name(
            lint_state,
            stanza_metadata,
            matched_stanzas,
            markdown_kind,
        )

    _info(
        f"Completion candidates: {[i.label for i in items] if items is not None else 'None'}"
    )

    return items


def deb822_hover(
    ls: "DebputyLanguageServer",
    params: HoverParams,
    file_metadata: Deb822FileMetadata[S, F],
    *,
    custom_handler: Optional[
        Callable[
            [
                "DebputyLanguageServer",
                Position,
                Optional[str],
                str,
                Optional[F],
                bool,
                "TextDocument",
                List[str],
            ],
            Optional[Hover],
        ]
    ] = None,
) -> Optional[Hover]:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lines = doc.lines
    deb822_file = ls.lint_state(doc).parsed_deb822_file_content
    if deb822_file is None:
        _warn("The deb822 result missing failed!?")
        ls.show_message_log(
            "Internal error; could not get deb822 content!?", MessageType.Warning
        )
        return None

    (
        server_pos,
        current_field,
        word_at_position,
        in_value,
        _,
        known_field,
        _,
    ) = _at_cursor(
        deb822_file,
        file_metadata,
        doc,
        lines,
        params.position,
    )

    if lines[server_pos.line].startswith("#"):
        return

    hover_text = None
    if custom_handler is not None:
        res = custom_handler(
            ls,
            server_pos,
            current_field,
            word_at_position,
            known_field,
            in_value,
            doc,
            lines,
        )
        if isinstance(res, Hover):
            return res
        hover_text = res

    if hover_text is None:
        if current_field is None:
            _info("No hover information as we cannot determine which field it is for")
            return None

        if known_field is None:
            return None
        if in_value:
            if not known_field.known_values:
                return None
            keyword = known_field.known_values.get(word_at_position)
            if keyword is None:
                return None
            hover_text = keyword.long_description_translated(ls)
            if hover_text is not None:
                header = "`{VALUE}` (Field: {FIELD_NAME})".format(
                    VALUE=keyword.value,
                    FIELD_NAME=known_field.name,
                )
                hover_text = f"# {header}\n\n{hover_text}"
        else:
            hover_text = known_field.long_description_translated(ls)
            if hover_text is None:
                hover_text = (
                    f"No documentation is available for the field {current_field}."
                )
            hover_text = f"# {known_field.name}\n\n{hover_text}"

    if hover_text is None:
        return None
    return Hover(
        contents=MarkupContent(
            kind=ls.hover_markup_format(MarkupKind.Markdown, MarkupKind.PlainText),
            value=hover_text,
        )
    )


def deb822_token_iter(
    tokens: Iterable[Deb822Token],
) -> Iterator[Tuple[Deb822Token, int, int, int, int]]:
    line_no = 0
    line_offset = 0

    for token in tokens:
        start_line = line_no
        start_line_offset = line_offset

        newlines = token.text.count("\n")
        line_no += newlines
        text_len = len(token.text)
        if newlines:
            if token.text.endswith("\n"):
                line_offset = 0
            else:
                # -2, one to remove the "\n" and one to get 0-offset
                line_offset = text_len - token.text.rindex("\n") - 2
        else:
            line_offset += text_len

        yield token, start_line, start_line_offset, line_no, line_offset


def deb822_folding_ranges(
    ls: "DebputyLanguageServer",
    params: FoldingRangeParams,
    # Unused for now: might be relevant for supporting folding for some fields
    _file_metadata: Deb822FileMetadata[Any, Any],
) -> Optional[Sequence[FoldingRange]]:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    comment_start = -1
    folding_ranges = []
    for (
        token,
        start_line,
        start_offset,
        end_line,
        end_offset,
    ) in deb822_token_iter(tokenize_deb822_file(doc.lines)):
        if token.is_comment:
            if comment_start < 0:
                comment_start = start_line
        elif comment_start > -1:
            comment_start = -1
            folding_range = FoldingRange(
                comment_start,
                end_line,
                kind=FoldingRangeKind.Comment,
            )

            folding_ranges.append(folding_range)

    return folding_ranges


class Deb822SemanticTokensState(SemanticTokensState):

    __slots__ = (
        "file_metadata",
        "keyword_token_code",
        "known_value_token_code",
        "comment_token_code",
        "substvars_token_code",
        "allow_overlapping_tokens",
    )

    def __init__(
        self,
        ls: "DebputyLanguageServer",
        doc: "TextDocument",
        lines: List[str],
        tokens: List[int],
        file_metadata: Deb822FileMetadata[Any, Any],
        keyword_token_code: int,
        known_value_token_code: int,
        comment_token_code: int,
        substvars_token_code: int,
    ) -> None:
        super().__init__(ls, doc, lines, tokens)
        self.file_metadata = file_metadata
        self.keyword_token_code = keyword_token_code
        self.known_value_token_code = known_value_token_code
        self.comment_token_code = comment_token_code
        self.substvars_token_code = substvars_token_code


def _emit_tokens_for_comment_element(
    sem_token_state: Deb822SemanticTokensState,
    comment_element: Deb822CommentElement,
    comment_start_line: int,
    comment_token_code: int,
) -> None:
    for comment_line_no, comment_token in enumerate(
        comment_element.iter_parts(),
        start=comment_start_line,
    ):
        assert comment_token.is_comment
        assert isinstance(comment_token, Deb822Token)
        sem_token_state.emit_token(
            Position(comment_line_no, 0),
            len(comment_token.text.rstrip()),
            comment_token_code,
        )


async def scan_for_syntax_errors_and_token_level_diagnostics(
    deb822_file: Deb822FileElement,
    lint_state: LintState,
) -> int:
    first_error = len(lint_state.lines) + 1
    spell_checker = lint_state.spellchecker()

    async for (
        token,
        start_line,
        start_offset,
        end_line,
        end_offset,
    ) in lint_state.slow_iter(deb822_token_iter(deb822_file.iter_tokens())):
        if token.is_error:
            first_error = min(first_error, start_line)
            token_range = TERange(
                TEPosition(
                    start_line,
                    start_offset,
                ),
                TEPosition(
                    end_line,
                    end_offset,
                ),
            )
            lint_state.emit_diagnostic(
                token_range,
                "Syntax error",
                "error",
                "debputy",
            )
        elif token.is_comment:
            for word, col_pos, end_col_pos in spell_checker.iter_words(token.text):
                corrections = spell_checker.provide_corrections_for(word)
                if not corrections:
                    continue
                word_range = TERange.between(
                    TEPosition(
                        start_line,
                        col_pos,
                    ),
                    TEPosition(
                        start_line,
                        end_col_pos,
                    ),
                )
                lint_state.emit_diagnostic(
                    word_range,
                    f'Spelling "{word}"',
                    "spelling",
                    "debputy",
                    quickfixes=[propose_correct_text_quick_fix(c) for c in corrections],
                    enable_non_interactive_auto_fix=False,
                )
    return first_error


async def _deb822_paragraph_semantic_tokens_full(
    ls: "DebputyLanguageServer",
    sem_token_state: Deb822SemanticTokensState,
    stanza: Deb822ParagraphElement,
    stanza_idx: int,
) -> None:
    doc = sem_token_state.doc
    keyword_token_code = sem_token_state.keyword_token_code
    known_value_token_code = sem_token_state.known_value_token_code
    comment_token_code = sem_token_state.comment_token_code

    stanza_position = stanza.position_in_file()
    stanza_metadata = sem_token_state.file_metadata.classify_stanza(
        stanza,
        stanza_idx=stanza_idx,
    )
    async for kvpair in ls.slow_iter(
        stanza.iter_parts_of_type(Deb822KeyValuePairElement), yield_every=25
    ):
        kvpair_position = kvpair.position_in_parent().relative_to(stanza_position)
        field_start = kvpair.field_token.position_in_parent().relative_to(
            kvpair_position
        )
        comment = kvpair.comment_element
        if comment:
            comment_start_line = field_start.line_position - len(comment)
            _emit_tokens_for_comment_element(
                sem_token_state,
                comment,
                comment_start_line,
                comment_token_code,
            )

        field_size = doc.position_codec.client_num_units(kvpair.field_name)

        sem_token_state.emit_token(
            te_position_to_lsp(field_start),
            field_size,
            keyword_token_code,
        )

        known_field: Optional[Deb822KnownField] = stanza_metadata.get(kvpair.field_name)
        if known_field is not None:
            if known_field.spellcheck_value:
                continue
            known_values: Container[str] = known_field.known_values or frozenset()
            interpretation = known_field.field_value_class.interpreter()
            field_disallows_substvars = (
                known_field.is_substvars_disabled_even_if_allowed_by_stanza
            )
            allow_substvars = (
                stanza_metadata.is_substvars_allowed_in_stanza
                and not field_disallows_substvars
            )
        else:
            known_values = frozenset()
            interpretation = None
            allow_substvars = stanza_metadata.is_substvars_allowed_in_stanza

        value_element_pos = kvpair.value_element.position_in_parent().relative_to(
            kvpair_position
        )
        if interpretation is None:
            for value_line in kvpair.value_element.value_lines:
                comment_element = value_line.comment_element
                if comment_element:
                    assert comment_element.position_in_parent().line_position == 0
                    comment_start_line = (
                        value_line.position_in_parent()
                        .relative_to(value_element_pos)
                        .line_position
                    )
                    _emit_tokens_for_comment_element(
                        sem_token_state,
                        comment_element,
                        comment_start_line,
                        comment_token_code,
                    )
            continue
        else:
            parts = kvpair.interpret_as(interpretation).iter_parts()
        for te in parts:
            if te.is_whitespace:
                continue
            if te.is_separator:
                continue
            value_range_in_parent_te = te.range_in_parent()
            value_range_te = value_range_in_parent_te.relative_to(value_element_pos)
            value = te.convert_to_text()
            if te.is_comment:
                token_type = comment_token_code
                value = value.rstrip()
            elif value in known_values:
                token_type = known_value_token_code
            elif allow_substvars and "${" in value:
                _process_value_with_substvars(
                    sem_token_state,
                    value,
                    value_range_te,
                    None,
                )
                continue
            else:
                continue
            value_len = doc.position_codec.client_num_units(value)
            sem_token_state.emit_token(
                te_position_to_lsp(value_range_te.start_pos),
                value_len,
                token_type,
            )


def _split_into_substvars(
    value: str,
    base_token_type: Optional[int],
    substvar_token_type: int,
) -> Iterable[Tuple[str, Optional[int]]]:

    i = 0
    next_search = i
    full_value_len = len(value)
    while i < full_value_len:
        try:
            subst_var_start = value.index("${", next_search)
            subst_var_end = value.index("}", subst_var_start + 2)
        except ValueError:
            token = value[i:full_value_len]
            if token:
                yield token, base_token_type
            return

        subst_var_end += 1
        subst_var = value[subst_var_start:subst_var_end]
        if subst_var != "${}" and not SUBSTVAR_RE.match(subst_var):
            subst_var = None

        if subst_var is None:
            next_search = subst_var_end
            continue

        token = value[i:subst_var_start]
        if token:
            yield token, base_token_type
        yield subst_var, substvar_token_type
        i = subst_var_end
        next_search = i


def _process_value_with_substvars(
    sem_token_state: Deb822SemanticTokensState,
    value: str,
    value_range_te: "TERange",
    base_token_type: Optional[int],
) -> None:
    pos_codec = sem_token_state.doc.position_codec

    # TODO: Support overlapping tokens if the editor does.

    line = value_range_te.start_pos.line_position
    token_pos = value_range_te.start_pos.cursor_position
    substvar_token_code = sem_token_state.substvars_token_code
    for token, token_type in _split_into_substvars(
        value,
        base_token_type,
        substvar_token_code,
    ):
        token_len = len(token)
        if token_type is not None:
            sem_token_state.emit_token(
                types.Position(line, token_pos),
                pos_codec.client_num_units(token),
                token_type,
            )
        token_pos += token_len


def deb822_format_file(
    lint_state: LintState,
    file_metadata: Deb822FileMetadata[Any, Any],
) -> Optional[Sequence[TextEdit]]:
    effective_preference = lint_state.effective_preference
    if effective_preference is None:
        return trim_end_of_line_whitespace(lint_state.position_codec, lint_state.lines)
    formatter = effective_preference.deb822_formatter()
    lines = lint_state.lines
    deb822_file = lint_state.parsed_deb822_file_content
    if deb822_file is None:
        _warn("The deb822 result missing failed!?")
        return None

    return list(
        file_metadata.reformat(
            effective_preference,
            deb822_file,
            formatter,
            lint_state.content,
            lint_state.position_codec,
            lines,
        )
    )


async def deb822_semantic_tokens_full(
    ls: "DebputyLanguageServer",
    request: SemanticTokensParams,
    file_metadata: Deb822FileMetadata[Any, Any],
) -> Optional[SemanticTokens]:
    doc = ls.workspace.get_text_document(request.text_document.uri)
    position_codec = doc.position_codec
    lines = doc.lines
    deb822_file = ls.lint_state(doc).parsed_deb822_file_content
    if deb822_file is None:
        _warn("The deb822 result missing failed!?")
        ls.show_message_log(
            "Internal error; could not get deb822 content!?", MessageType.Warning
        )
        return None

    tokens: List[int] = []
    comment_token_code = SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Comment.value]
    sem_token_state = Deb822SemanticTokensState(
        ls,
        doc,
        lines,
        tokens,
        file_metadata,
        SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Keyword],
        SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.EnumMember],
        comment_token_code,
        SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Macro],
    )

    stanza_idx = 0

    async for part in ls.slow_iter(deb822_file.iter_parts(), yield_every=20):
        if part.is_comment:
            pos = part.position_in_file()
            sem_token_state.emit_token(
                te_position_to_lsp(pos),
                # Avoid trailing newline
                position_codec.client_num_units(part.convert_to_text().rstrip()),
                comment_token_code,
            )
        elif isinstance(part, Deb822ParagraphElement):
            await _deb822_paragraph_semantic_tokens_full(
                ls,
                sem_token_state,
                part,
                stanza_idx,
            )
            stanza_idx += 1
    if not tokens:
        return None
    return SemanticTokens(tokens)


def _complete_field_name(
    lint_state: LintState,
    stanza_metadata: StanzaMetadata[Any],
    matched_stanzas: Iterable[Deb822ParagraphElement],
    markdown_kind: MarkupKind,
) -> Sequence[CompletionItem]:
    items = []
    matched_stanzas = list(matched_stanzas)
    seen_fields = set(
        stanza_metadata.normalize_field_name(f.lower())
        for f in chain.from_iterable(
            # The typing from python3-debian is not entirely optimal here. The iter always return a
            # `str`, but the provided type is `ParagraphKey` (because `__getitem__` supports those)
            # and that is not exclusively a `str`.
            #
            # So, this cast for now
            cast("Iterable[str]", s)
            for s in matched_stanzas
        )
    )
    for cand_key, cand in stanza_metadata.items():
        if stanza_metadata.normalize_field_name(cand_key.lower()) in seen_fields:
            continue
        item = cand.complete_field(lint_state, matched_stanzas, markdown_kind)
        if item is not None:
            items.append(item)
    return items



import dataclasses

from debputy.manifest_parser.base_types import FileSystemMode


@dataclasses.dataclass(slots=True)
class DebputyCapability:
    capabilities: str
    capability_mode: FileSystemMode
    definition_source: str



from dataclasses import dataclass

@dataclass
class Library:
    ESC: str = ...
    END: str = ...
    FOREGROUND_256: str = ...
    BACKGROUND_256: str = ...
    FOREGROUND_RGB: str = ...
    BACKGROUND_RGB: str = ...
    UNDERLINE_COLOR: str = ...
    CONTROLS = ...
    STYLES = ...
    COLORTERM = ...
    COLORS = ...
    HEX_COLORS = ...
    def __init__(self, ESC, END, FOREGROUND_256, BACKGROUND_256, FOREGROUND_RGB, BACKGROUND_RGB, UNDERLINE_COLOR) -> None: ...



import dataclasses
import textwrap
from typing import Mapping, Dict, Self

import pytest
from debian.deb822 import Deb822
from debian.debian_support import DpkgArchTable

from debputy._deb_options_profiles import DebBuildOptionsAndProfiles
from debputy.architecture_support import DpkgArchitectureBuildProcessValuesTable
from debputy.highlevel_manifest import (
    HighLevelManifest,
    PathNotCoveredByInstallRulesError,
)
from debputy.highlevel_manifest_parser import YAMLManifestParser
from debputy.installations import (
    InstallSearchDirContext,
    NoMatchForInstallPatternError,
    SearchDir,
)
from debputy.packages import BinaryPackage, SourcePackage
from debputy.plugin.api import virtual_path_def, VirtualPath
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.plugin.api.spec import INTEGRATION_MODE_DH_DEBPUTY
from debputy.plugin.api.test_api import build_virtual_file_system
from debputy.substitution import Substitution


@pytest.fixture()
def manifest_parser_pkg_foo(
    amd64_dpkg_architecture_variables,
    dpkg_arch_query,
    source_package,
    package_single_foo_arch_all_cxt_amd64,
    amd64_substitution,
    no_profiles_or_build_options,
    debputy_plugin_feature_set,
) -> YAMLManifestParser:
    # We need an empty directory to avoid triggering packager provided files.
    debian_dir = build_virtual_file_system([])
    return YAMLManifestParser(
        "debian/test-debputy.manifest",
        source_package,
        package_single_foo_arch_all_cxt_amd64,
        amd64_substitution,
        amd64_dpkg_architecture_variables,
        dpkg_arch_query,
        no_profiles_or_build_options,
        debputy_plugin_feature_set,
        "full",
        debian_dir=debian_dir,
    )


@pytest.fixture()
def manifest_parser_pkg_foo_w_udeb(
    amd64_dpkg_architecture_variables,
    dpkg_arch_query,
    source_package,
    package_foo_w_udeb_arch_any_cxt_amd64,
    amd64_substitution,
    no_profiles_or_build_options,
    debputy_plugin_feature_set,
) -> YAMLManifestParser:
    # We need an empty directory to avoid triggering packager provided files.
    debian_dir = build_virtual_file_system([])
    return YAMLManifestParser(
        "debian/test-debputy.manifest",
        source_package,
        package_foo_w_udeb_arch_any_cxt_amd64,
        amd64_substitution,
        amd64_dpkg_architecture_variables,
        dpkg_arch_query,
        no_profiles_or_build_options,
        debputy_plugin_feature_set,
        "full",
        debian_dir=debian_dir,
    )


@dataclasses.dataclass(slots=True, frozen=True)
class FlexibleManifestParser:
    dpkg_architecture_variables: DpkgArchitectureBuildProcessValuesTable
    dpkg_arch_query: DpkgArchTable
    source_package: SourcePackage
    substitution: Substitution
    no_profiles_or_build_options: DebBuildOptionsAndProfiles
    debputy_plugin_feature_set: PluginProvidedFeatureSet
    debian_dir: VirtualPath
    binary_packages: Dict[str, BinaryPackage] = dataclasses.field(default_factory=dict)

    def binary_package(
        self,
        fields: Mapping[str, str],
        *,
        is_main_package: bool = False,
        should_be_acted_on: bool = True,
    ) -> Self:
        pkg = BinaryPackage(
            fields,
            self.dpkg_architecture_variables,
            self.dpkg_arch_query,
            is_main_package=is_main_package,
            should_be_acted_on=should_be_acted_on,
        )
        self.binary_packages[pkg.name] = pkg
        return self

    def parse_manifest(self, manifest_data: str) -> HighLevelManifest:
        binary_packages = self.binary_packages
        if not binary_packages:
            raise ValueError("No binary packages defined!?")
        return YAMLManifestParser(
            "debian/test-debputy.manifest",
            self.source_package,
            binary_packages,
            self.substitution,
            self.dpkg_architecture_variables,
            self.dpkg_arch_query,
            self.no_profiles_or_build_options,
            self.debputy_plugin_feature_set,
            "full",
            debian_dir=self.debian_dir,
        ).parse_manifest(fd=manifest_data)


@pytest.fixture()
def flexible_manifest_parser(
    amd64_dpkg_architecture_variables: DpkgArchitectureBuildProcessValuesTable,
    dpkg_arch_query: DpkgArchTable,
    source_package: SourcePackage,
    amd64_substitution: Substitution,
    no_profiles_or_build_options: DebBuildOptionsAndProfiles,
    debputy_plugin_feature_set: PluginProvidedFeatureSet,
) -> FlexibleManifestParser:
    return FlexibleManifestParser(
        amd64_dpkg_architecture_variables,
        dpkg_arch_query,
        source_package,
        amd64_substitution,
        no_profiles_or_build_options,
        debputy_plugin_feature_set,
        build_virtual_file_system([]),
    )


def test_install_rules(manifest_parser_pkg_foo) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [virtual_path_def(".", fs_path="/nowhere")]
    )
    debian_tmp_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/tmp"),
            virtual_path_def("usr/", fs_path="/nowhere/debian/tmp/usr"),
            virtual_path_def("usr/bin/", fs_path="/nowhere/debian/tmp/usr/bin"),
            virtual_path_def(
                "usr/bin/foo",
                fs_path="/nowhere/debian/tmp/usr/bin/foo",
                content="#!/bin/sh\n",
                mtime=10,
            ),
            virtual_path_def(
                "usr/bin/foo-util",
                fs_path="/nowhere/debian/tmp/usr/bin/foo-util",
                content="#!/bin/sh\n",
                mtime=10,
            ),
            virtual_path_def(
                "usr/bin/tool.sh",
                fs_path="/nowhere/debian/tmp/usr/bin/tool.sh",
                link_target="./foo",
            ),
            virtual_path_def("usr/share/", fs_path="/nowhere/debian/tmp/usr/share"),
            virtual_path_def(
                "usr/share/foo/", fs_path="/nowhere/debian/tmp/usr/share/foo"
            ),
            virtual_path_def(
                "usr/share/foo/foo.txt",
                fs_path="/nowhere/debian/tmp/usr/share/foo/foo.txt",
                content="A text file",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              - /usr/share/foo
              - /usr/bin/foo
              - /usr/bin/foo-util
          - install:
              source: usr/bin/tool.sh
              as: usr/bin/tool
        """
    )
    manifest = manifest_parser_pkg_foo.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_tmp_dir, all_pkgs),
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [debian_tmp_dir],
        ),
    )
    assert "foo" in result
    foo_fs_root = result["foo"].fs_root

    ub_dir = foo_fs_root.lookup("/usr/bin")
    assert ub_dir is not None
    assert ub_dir.is_dir
    assert not ub_dir.has_fs_path  # This will be "generated"

    tool = ub_dir.get("tool")
    assert tool is not None
    assert tool.is_symlink
    assert tool.readlink() == "./foo"

    assert {"foo", "foo-util", "tool"} == {p.name for p in ub_dir.iterdir}
    for n in ["foo", "foo-util"]:
        assert ub_dir[n].mtime == 10
    usf_dir = foo_fs_root.lookup("/usr/share/foo")
    assert usf_dir is not None
    assert usf_dir.is_dir
    # Here we are installing an actual directory, so it should be present too
    assert usf_dir.has_fs_path
    assert usf_dir.fs_path == "/nowhere/debian/tmp/usr/share/foo"
    assert {"foo.txt"} == {p.name for p in usf_dir.iterdir}


def test_multi_dest_install_rules(manifest_parser_pkg_foo) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere"),
            virtual_path_def("source/", fs_path="/nowhere/source"),
            virtual_path_def("source/foo/", fs_path="/nowhere/foo"),
            virtual_path_def(
                "source/foo/foo-a.data",
                fs_path="/nowhere/foo/foo-a.data",
                content="data file",
            ),
            virtual_path_def(
                "source/foo/foo-b.data",
                fs_path="/nowhere/foo/foo-b.data",
                link_target="./foo-a.data",
            ),
            virtual_path_def("source/bar/", fs_path="/nowhere/bar"),
            virtual_path_def(
                "source/bar/bar-a.data",
                fs_path="/nowhere/bar/bar-a.data",
                content="data file",
            ),
            virtual_path_def(
                "source/bar/bar-b.data",
                fs_path="/nowhere/bar/bar-b.data",
                content="data file",
            ),
            virtual_path_def(
                "source/tool.sh",
                fs_path="/nowhere/source/tool.sh",
                content="#!/bin/sh\n# run some command ...",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - multi-dest-install:
              sources:
              - source/foo/*
              - source/bar
              dest-dirs:
              - usr/share/foo
              - usr/share/foo2
          - multi-dest-install:
              source: source/tool.sh
              as:
               - usr/share/foo/tool
               - usr/share/foo2/tool
        """
    )
    manifest = manifest_parser_pkg_foo.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [],
        ),
    )
    assert "foo" in result
    foo_fs_root = result["foo"].fs_root

    for stem in ["foo", "foo2"]:
        foo_dir = foo_fs_root.lookup(f"/usr/share/{stem}")
        assert foo_dir is not None
        assert foo_dir.is_dir

        assert {"foo-a.data", "foo-b.data", "bar", "tool"} == {
            p.name for p in foo_dir.iterdir
        }

        tool = foo_dir["tool"]
        assert tool.is_file
        with tool.open() as fd:
            content = fd.read()
        assert content.startswith("#!/bin/sh")
        foo_a = foo_dir["foo-a.data"]
        assert foo_a.is_file
        assert foo_a.fs_path == "/nowhere/foo/foo-a.data"
        with foo_a.open() as fd:
            content = fd.read()
        assert "data" in content
        foo_b = foo_dir["foo-b.data"]
        assert foo_b.is_symlink
        assert foo_b.readlink() == "./foo-a.data"

        bar = foo_dir["bar"]
        assert bar.is_dir
        assert {"bar-a.data", "bar-b.data"} == {p.name for p in bar.iterdir}
        assert {"/nowhere/bar/bar-a.data", "/nowhere/bar/bar-b.data"} == {
            p.fs_path for p in bar.iterdir
        }


def test_install_rules_with_glob(manifest_parser_pkg_foo) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [virtual_path_def(".", fs_path="/nowhere")]
    )
    debian_tmp_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/tmp"),
            virtual_path_def("usr/", fs_path="/nowhere/debian/tmp/usr"),
            virtual_path_def("usr/bin/", fs_path="/nowhere/debian/tmp/usr/bin"),
            virtual_path_def(
                "usr/bin/foo",
                fs_path="/nowhere/debian/tmp/usr/bin/foo",
                content="#!/bin/sh\n",
            ),
            virtual_path_def(
                "usr/bin/foo-util",
                fs_path="/nowhere/debian/tmp/usr/bin/foo-util",
                content="#!/bin/sh\n",
            ),
            virtual_path_def(
                "usr/bin/tool.sh",
                fs_path="/nowhere/debian/tmp/usr/bin/tool.sh",
                link_target="./foo",
            ),
            virtual_path_def("usr/share/", fs_path="/nowhere/debian/tmp/usr/share"),
            virtual_path_def(
                "usr/share/foo/", fs_path="/nowhere/debian/tmp/usr/share/foo"
            ),
            virtual_path_def(
                "usr/share/foo/foo.txt",
                fs_path="/nowhere/debian/tmp/usr/share/foo/foo.txt",
                content="A text file",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              source: usr/bin/tool.sh
              as: usr/bin/tool
          - install:
              - /usr/share/foo
              - /usr/bin/foo*
        """
    )
    manifest = manifest_parser_pkg_foo.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_tmp_dir, all_pkgs),
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [debian_tmp_dir],
        ),
    )
    assert "foo" in result
    foo_fs_root = result["foo"].fs_root

    ub_dir = foo_fs_root.lookup("/usr/bin")
    assert ub_dir is not None
    assert ub_dir.is_dir
    assert not ub_dir.has_fs_path  # This will be "generated"

    tool = ub_dir.get("tool")
    assert tool is not None
    assert tool.is_symlink
    assert tool.readlink() == "./foo"

    assert {"foo", "foo-util", "tool"} == {p.name for p in ub_dir.iterdir}
    usf_dir = foo_fs_root.lookup("/usr/share/foo")
    assert usf_dir is not None
    assert usf_dir.is_dir
    # Here we are installing an actual directory, so it should be present too
    assert usf_dir.has_fs_path
    assert usf_dir.fs_path == "/nowhere/debian/tmp/usr/share/foo"
    assert {"foo.txt"} == {p.name for p in usf_dir.iterdir}


def test_install_rules_auto_discard_rules_dir(manifest_parser_pkg_foo) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [virtual_path_def(".", fs_path="/nowhere")]
    )
    debian_tmp_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/tmp"),
            virtual_path_def("usr/", fs_path="/nowhere/debian/tmp/usr"),
            virtual_path_def("usr/lib/", fs_path="/nowhere/debian/tmp/usr/lib"),
            virtual_path_def(
                "usr/lib/libfoo.so.1.0.0",
                fs_path="/nowhere/debian/tmp/usr/lib/libfoo.so.1.0.0",
                content="Not really an ELF FILE",
            ),
            virtual_path_def(
                "usr/lib/libfoo.la",
                fs_path="/nowhere/debian/tmp/usr/lib/libfoo.la",
                content="Not really a LA FILE",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              - /usr/lib
        """
    )
    manifest = manifest_parser_pkg_foo.parse_manifest(fd=manifest_content)

    all_pkgs = frozenset(manifest.all_packages)

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_tmp_dir, all_pkgs),
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [debian_tmp_dir],
        ),
    )
    assert "foo" in result
    foo_fs_root = result["foo"].fs_root

    lib_dir = foo_fs_root.lookup("/usr/lib")
    assert lib_dir is not None
    assert lib_dir.is_dir
    assert lib_dir.has_fs_path
    assert lib_dir.fs_path == "/nowhere/debian/tmp/usr/lib"

    so_file = lib_dir.get("libfoo.so.1.0.0")
    assert so_file is not None
    assert so_file.is_file
    assert so_file.has_fs_path
    assert so_file.fs_path == "/nowhere/debian/tmp/usr/lib/libfoo.so.1.0.0"

    assert {"libfoo.so.1.0.0"} == {p.name for p in lib_dir.iterdir}


def test_install_rules_auto_discard_rules_glob(manifest_parser_pkg_foo) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [virtual_path_def(".", fs_path="/nowhere")]
    )
    debian_tmp_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/tmp"),
            virtual_path_def("usr/", fs_path="/nowhere/debian/tmp/usr"),
            virtual_path_def("usr/lib/", fs_path="/nowhere/debian/tmp/usr/lib"),
            virtual_path_def(
                "usr/lib/libfoo.so.1.0.0",
                fs_path="/nowhere/debian/tmp/usr/lib/libfoo.so.1.0.0",
                content="Not really an ELF FILE",
            ),
            virtual_path_def(
                "usr/lib/libfoo.la",
                fs_path="/nowhere/debian/tmp/usr/lib/libfoo.la",
                content="Not really an ELF FILE",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              - /usr/lib/*
        """
    )
    manifest = manifest_parser_pkg_foo.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_tmp_dir, all_pkgs),
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [debian_tmp_dir],
        ),
    )
    assert "foo" in result
    foo_fs_root = result["foo"].fs_root

    lib_dir = foo_fs_root.lookup("/usr/lib")
    assert lib_dir is not None
    assert lib_dir.is_dir
    assert not lib_dir.has_fs_path

    so_file = lib_dir.get("libfoo.so.1.0.0")
    assert so_file is not None
    assert so_file.is_file
    assert so_file.has_fs_path
    assert so_file.fs_path == "/nowhere/debian/tmp/usr/lib/libfoo.so.1.0.0"

    assert {"libfoo.so.1.0.0"} == {p.name for p in lib_dir.iterdir}


def test_install_rules_auto_discard_rules_overruled_by_explicit_install_rule(
    manifest_parser_pkg_foo,
) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [virtual_path_def(".", fs_path="/nowhere")]
    )
    debian_tmp_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/tmp"),
            virtual_path_def("usr/", fs_path="/nowhere/debian/tmp/usr"),
            virtual_path_def("usr/lib/", fs_path="/nowhere/debian/tmp/usr/lib"),
            virtual_path_def(
                "usr/lib/libfoo.so.1.0.0",
                fs_path="/nowhere/debian/tmp/usr/lib/libfoo.so.1.0.0",
                content="Not really an ELF FILE",
            ),
            virtual_path_def(
                "usr/lib/libfoo.la",
                fs_path="/nowhere/debian/tmp/usr/lib/libfoo.la",
                content="Not really an ELF FILE",
            ),
            virtual_path_def(
                "usr/lib/libfoo.so",
                fs_path="/nowhere/debian/tmp/usr/lib/libfoo.so",
                link_target="libfoo.so.1.0.0",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              - /usr/lib
              - /usr/lib/libfoo.la
        """
    )
    manifest = manifest_parser_pkg_foo.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_tmp_dir, all_pkgs),
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [debian_tmp_dir],
        ),
    )
    assert "foo" in result
    foo_fs_root = result["foo"].fs_root

    lib_dir = foo_fs_root.lookup("/usr/lib")
    assert lib_dir is not None
    assert lib_dir.is_dir
    assert lib_dir.has_fs_path
    assert lib_dir.fs_path == "/nowhere/debian/tmp/usr/lib"

    so_file = lib_dir.get("libfoo.so.1.0.0")
    assert so_file is not None
    assert so_file.is_file
    assert so_file.has_fs_path
    assert so_file.fs_path == "/nowhere/debian/tmp/usr/lib/libfoo.so.1.0.0"

    la_file = lib_dir.get("libfoo.la")
    assert la_file is not None
    assert la_file.is_file
    assert la_file.has_fs_path
    assert la_file.fs_path == "/nowhere/debian/tmp/usr/lib/libfoo.la"

    so_link = lib_dir.get("libfoo.so")
    assert so_link is not None
    assert so_link.is_symlink
    assert so_link.readlink() == "libfoo.so.1.0.0"

    assert {"libfoo.so.1.0.0", "libfoo.so", "libfoo.la"} == {
        p.name for p in lib_dir.iterdir
    }


def test_install_rules_install_as_with_var(manifest_parser_pkg_foo) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere"),
            virtual_path_def("build/", fs_path="/nowhere/build"),
            virtual_path_def(
                "build/private-arch-tool.sh",
                fs_path="/nowhere/build/private-arch-tool.sh",
                content="#!/bin/sh\n",
            ),
        ]
    )
    debian_tmp_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/tmp"),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              source: build/private-arch-tool.sh
              as: /usr/lib/{{DEB_HOST_MULTIARCH}}/foo/private-arch-tool
        """
    )
    manifest = manifest_parser_pkg_foo.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_tmp_dir, all_pkgs),
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [debian_tmp_dir],
        ),
    )
    assert "foo" in result
    foo_fs_root = result["foo"].fs_root

    # The variable is always resolved in amd64 context, so we can hard code the resolved
    # variable
    tool = foo_fs_root.lookup("/usr/lib/x86_64-linux-gnu/foo/private-arch-tool")
    assert tool is not None
    assert tool.is_file
    assert tool.fs_path == "/nowhere/build/private-arch-tool.sh"


def test_install_rules_no_matches(manifest_parser_pkg_foo) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere"),
            virtual_path_def("build/", fs_path="/nowhere/build"),
            virtual_path_def(
                "build/private-arch-tool.sh",
                fs_path="/nowhere/build/private-arch-tool.sh",
                content="#!/bin/sh\n",
            ),
        ]
    )
    debian_tmp_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/tmp"),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              # Typo: the path should have ended with ".sh"
              source: build/private-arch-tool
              as: /usr/lib/foo/private-arch-tool
        """
    )
    manifest = manifest_parser_pkg_foo.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    with pytest.raises(NoMatchForInstallPatternError) as e_info:
        manifest.perform_installations(
            INTEGRATION_MODE_DH_DEBPUTY,
            install_request_context=InstallSearchDirContext(
                [
                    SearchDir(debian_tmp_dir, all_pkgs),
                    SearchDir(debian_source_root_dir, all_pkgs),
                ],
                [debian_tmp_dir],
            ),
        )
    expected_msg = (
        "There were no matches for build/private-arch-tool in /nowhere/debian/tmp, /nowhere"
        " (definition: installations[0].install [Line 5 column 6])."
        " Match rule: ./build/private-arch-tool (the exact path / no globbing)"
    )
    assert e_info.value.message == expected_msg


def test_install_rules_per_package_search_dirs(manifest_parser_pkg_foo_w_udeb) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [virtual_path_def(".", fs_path="/nowhere")]
    )
    debian_tmp_deb_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/tmp-deb"),
            virtual_path_def("usr/", fs_path="/nowhere/debian/tmp-deb/usr"),
            virtual_path_def("usr/bin/", fs_path="/nowhere/debian/tmp-deb/usr/bin"),
            virtual_path_def(
                "usr/bin/foo",
                fs_path="/nowhere/debian/tmp-deb/usr/bin/foo",
                content="#!/bin/sh\ndeb",
            ),
            virtual_path_def(
                "usr/bin/foo-util",
                fs_path="/nowhere/debian/tmp-deb/usr/bin/foo-util",
                content="#!/bin/sh\ndeb",
            ),
            virtual_path_def(
                "usr/bin/tool.sh",
                fs_path="/nowhere/debian/tmp-deb/usr/bin/tool.sh",
                link_target="./foo",
            ),
            virtual_path_def("usr/share/", fs_path="/nowhere/debian/tmp-deb/usr/share"),
            virtual_path_def(
                "usr/share/foo/", fs_path="/nowhere/debian/tmp-deb/usr/share/foo"
            ),
            virtual_path_def(
                "usr/share/foo/foo.txt",
                fs_path="/nowhere/debian/tmp-deb/usr/share/foo/foo.txt",
                content="A deb text file",
            ),
        ]
    )
    debian_tmp_udeb_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/tmp-udeb"),
            virtual_path_def("usr/", fs_path="/nowhere/debian/tmp-udeb/usr"),
            virtual_path_def("usr/bin/", fs_path="/nowhere/debian/tmp-udeb/usr/bin"),
            virtual_path_def(
                "usr/bin/foo",
                fs_path="/nowhere/debian/tmp-udeb/usr/bin/foo",
                content="#!/bin/sh\nudeb",
            ),
            virtual_path_def(
                "usr/bin/foo-util",
                fs_path="/nowhere/debian/tmp-udeb/usr/bin/foo-util",
                content="#!/bin/sh\nudeb",
            ),
            virtual_path_def(
                "usr/bin/tool.sh",
                fs_path="/nowhere/debian/tmp-udeb/usr/bin/tool.sh",
                link_target="./foo",
            ),
            virtual_path_def(
                "usr/share/", fs_path="/nowhere/debian/tmp-udeb/usr/share"
            ),
            virtual_path_def(
                "usr/share/foo/", fs_path="/nowhere/debian/tmp-udeb/usr/share/foo"
            ),
            virtual_path_def(
                "usr/share/foo/foo.txt",
                fs_path="/nowhere/debian/tmp-udeb/usr/share/foo/foo.txt",
                content="A udeb text file",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              source: usr/bin/tool.sh
              as: usr/bin/tool
              into:
                - foo
                - foo-udeb
          - install:
              sources:
                - /usr/share/foo
                - /usr/bin/foo*
              into:
                - foo
                - foo-udeb
        """
    )
    manifest = manifest_parser_pkg_foo_w_udeb.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)
    all_deb_pkgs = frozenset({p for p in all_pkgs if not p.is_udeb})
    all_udeb_pkgs = frozenset({p for p in all_pkgs if p.is_udeb})

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_tmp_deb_dir, all_deb_pkgs),
                SearchDir(debian_tmp_udeb_dir, all_udeb_pkgs),
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [debian_tmp_deb_dir],
        ),
    )
    for pkg, ptype in [
        ("foo", "deb"),
        ("foo-udeb", "udeb"),
    ]:
        assert pkg in result
        fs_root = result[pkg].fs_root

        ub_dir = fs_root.lookup("/usr/bin")
        assert ub_dir is not None
        assert ub_dir.is_dir
        assert not ub_dir.has_fs_path  # This will be "generated"

        tool = ub_dir.get("tool")
        assert tool is not None
        assert tool.is_symlink
        assert tool.readlink() == "./foo"

        assert {"foo", "foo-util", "tool"} == {p.name for p in ub_dir.iterdir}

        for p in ub_dir.iterdir:
            assert p.has_fs_path
            assert f"/nowhere/debian/tmp-{ptype}/usr/bin" in p.fs_path

        usf_dir = fs_root.lookup("/usr/share/foo")
        assert usf_dir is not None
        assert usf_dir.is_dir
        # Here we are installing an actual directory, so it should be present too
        assert usf_dir.has_fs_path
        assert usf_dir.fs_path == f"/nowhere/debian/tmp-{ptype}/usr/share/foo"
        assert {"foo.txt"} == {p.name for p in usf_dir.iterdir}
        foo_txt = usf_dir["foo.txt"]
        assert foo_txt.fs_path == f"/nowhere/debian/tmp-{ptype}/usr/share/foo/foo.txt"


def test_install_rules_multi_into(manifest_parser_pkg_foo_w_udeb) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere"),
            virtual_path_def("source/", fs_path="/nowhere/source"),
            virtual_path_def("source/foo/", fs_path="/nowhere/foo"),
            virtual_path_def(
                "source/foo/foo-a.data",
                fs_path="/nowhere/foo/foo-a.data",
                content="data file",
            ),
            virtual_path_def(
                "source/foo/foo-b.data",
                fs_path="/nowhere/foo/foo-b.data",
                link_target="./foo-a.data",
            ),
            virtual_path_def("source/bar/", fs_path="/nowhere/bar"),
            virtual_path_def(
                "source/bar/bar-a.data",
                fs_path="/nowhere/bar/bar-a.data",
                content="data file",
            ),
            virtual_path_def(
                "source/bar/bar-b.data",
                fs_path="/nowhere/bar/bar-b.data",
                content="data file",
            ),
            virtual_path_def(
                "source/tool.sh",
                fs_path="/nowhere/source/tool.sh",
                content="#!/bin/sh\n# run some command ...",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              sources:
              - source/foo/*
              - source/bar
              dest-dir: usr/share/foo
              into:
              - foo
              - foo-udeb
          - install:
              source: source/tool.sh
              as: usr/share/foo/tool
              into:
              - foo
              - foo-udeb
        """
    )
    manifest = manifest_parser_pkg_foo_w_udeb.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [],
        ),
    )
    for pkg in ["foo", "foo-udeb"]:
        assert pkg in result
        foo_fs_root = result[pkg].fs_root

        foo_dir = foo_fs_root.lookup("/usr/share/foo")
        assert foo_dir is not None
        assert foo_dir.is_dir

        assert {"foo-a.data", "foo-b.data", "bar", "tool"} == {
            p.name for p in foo_dir.iterdir
        }

        tool = foo_dir["tool"]
        assert tool.is_file
        with tool.open() as fd:
            content = fd.read()
        assert content.startswith("#!/bin/sh")
        foo_a = foo_dir["foo-a.data"]
        assert foo_a.is_file
        assert foo_a.fs_path == "/nowhere/foo/foo-a.data"
        with foo_a.open() as fd:
            content = fd.read()
        assert "data" in content
        foo_b = foo_dir["foo-b.data"]
        assert foo_b.is_symlink
        assert foo_b.readlink() == "./foo-a.data"

        bar = foo_dir["bar"]
        assert bar.is_dir
        assert {"bar-a.data", "bar-b.data"} == {p.name for p in bar.iterdir}
        assert {"/nowhere/bar/bar-a.data", "/nowhere/bar/bar-b.data"} == {
            p.fs_path for p in bar.iterdir
        }


def test_auto_install_d_pkg(manifest_parser_pkg_foo_w_udeb) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [virtual_path_def(".", fs_path="/nowhere")]
    )
    debian_foo_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/foo"),
            virtual_path_def("usr/", fs_path="/nowhere/debian/foo/usr"),
            virtual_path_def("usr/bin/", fs_path="/nowhere/debian/foo/usr/bin"),
            virtual_path_def(
                "usr/bin/foo",
                fs_path="/nowhere/debian/foo/usr/bin/foo",
                content="#!/bin/sh\ndeb",
            ),
            virtual_path_def(
                "usr/bin/foo-util",
                fs_path="/nowhere/debian/foo/usr/bin/foo-util",
                content="#!/bin/sh\ndeb",
            ),
            virtual_path_def(
                "usr/bin/tool",
                fs_path="/nowhere/debian/foo/usr/bin/tool",
                link_target="./foo",
            ),
            virtual_path_def("usr/share/", fs_path="/nowhere/debian/foo/usr/share"),
            virtual_path_def(
                "usr/share/foo/", fs_path="/nowhere/debian/foo/usr/share/foo"
            ),
            virtual_path_def(
                "usr/share/foo/foo.txt",
                fs_path="/nowhere/debian/foo/usr/share/foo/foo.txt",
                content="A deb text file",
            ),
        ]
    )
    debian_foo_udeb_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere/debian/foo-udeb"),
            virtual_path_def("usr/", fs_path="/nowhere/debian/foo-udeb/usr"),
            virtual_path_def("usr/bin/", fs_path="/nowhere/debian/foo-udeb/usr/bin"),
            virtual_path_def(
                "usr/bin/foo",
                fs_path="/nowhere/debian/foo-udeb/usr/bin/foo",
                content="#!/bin/sh\nudeb",
            ),
            virtual_path_def(
                "usr/bin/foo-util",
                fs_path="/nowhere/debian/foo-udeb/usr/bin/foo-util",
                content="#!/bin/sh\nudeb",
            ),
            virtual_path_def(
                "usr/bin/tool",
                fs_path="/nowhere/debian/foo-udeb/usr/bin/tool",
                link_target="./foo",
            ),
            virtual_path_def(
                "usr/share/", fs_path="/nowhere/debian/foo-udeb/usr/share"
            ),
            virtual_path_def(
                "usr/share/foo/", fs_path="/nowhere/debian/foo-udeb/usr/share/foo"
            ),
            virtual_path_def(
                "usr/share/foo/foo.txt",
                fs_path="/nowhere/debian/foo-udeb/usr/share/foo/foo.txt",
                content="A udeb text file",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        """
    )
    manifest = manifest_parser_pkg_foo_w_udeb.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [debian_foo_dir],
            {
                "foo": debian_foo_dir,
                "foo-udeb": debian_foo_udeb_dir,
            },
        ),
    )
    for pkg in ["foo", "foo-udeb"]:
        assert pkg in result
        fs_root = result[pkg].fs_root
        ub_dir = fs_root.lookup("/usr/bin")
        assert ub_dir is not None
        assert ub_dir.is_dir
        assert ub_dir.has_fs_path
        assert ub_dir.fs_path == f"/nowhere/debian/{pkg}/usr/bin"

        assert {"foo", "foo-util", "tool"} == {p.name for p in ub_dir.iterdir}

        tool = ub_dir.get("tool")
        assert tool is not None
        assert tool.is_symlink
        assert tool.readlink() == "./foo"

        for p in ub_dir.iterdir:
            assert p.has_fs_path
            assert f"/nowhere/debian/{pkg}/usr/bin" in p.fs_path

        usf_dir = fs_root.lookup("/usr/share/foo")
        assert usf_dir is not None
        assert usf_dir.is_dir
        # Here we are installing an actual directory, so it should be present too
        assert usf_dir.has_fs_path
        assert usf_dir.fs_path == f"/nowhere/debian/{pkg}/usr/share/foo"
        assert {"foo.txt"} == {p.name for p in usf_dir.iterdir}
        foo_txt = usf_dir["foo.txt"]
        assert foo_txt.fs_path == f"/nowhere/debian/{pkg}/usr/share/foo/foo.txt"


def test_install_doc_rules_ignore_udeb(manifest_parser_pkg_foo_w_udeb) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere"),
            virtual_path_def("source/", fs_path="/nowhere/source"),
            virtual_path_def("source/foo/", fs_path="/nowhere/foo"),
            virtual_path_def(
                "source/foo/foo-a.txt",
                fs_path="/nowhere/foo/foo-a.txt",
                content="data file",
            ),
            virtual_path_def(
                "source/foo/foo-b.txt",
                fs_path="/nowhere/foo/foo-b.txt",
                link_target="./foo-a.txt",
            ),
            virtual_path_def("source/html/", fs_path="/nowhere/html"),
            virtual_path_def(
                "source/html/bar-a.html",
                fs_path="/nowhere/html/bar-a.html",
                content="data file",
            ),
            virtual_path_def(
                "source/html/bar-b.html",
                fs_path="/nowhere/html/bar-b.html",
                content="data file",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install-doc:
              sources:
              - source/foo/*
              - source/html
        """
    )
    manifest = manifest_parser_pkg_foo_w_udeb.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    result = manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [],
        ),
    )
    assert "foo" in result
    foo_fs_root = result["foo"].fs_root

    foo_dir = foo_fs_root.lookup("/usr/share/doc/foo")
    assert foo_dir is not None
    assert foo_dir.is_dir

    assert {"foo-a.txt", "foo-b.txt", "html"} == {p.name for p in foo_dir.iterdir}

    foo_a = foo_dir["foo-a.txt"]
    assert foo_a.is_file
    assert foo_a.fs_path == "/nowhere/foo/foo-a.txt"
    foo_b = foo_dir["foo-b.txt"]
    assert foo_b.is_symlink
    assert foo_b.readlink() == "./foo-a.txt"

    html_dir = foo_dir["html"]
    assert html_dir.is_dir
    assert {"bar-a.html", "bar-b.html"} == {p.name for p in html_dir.iterdir}
    assert {"/nowhere/html/bar-a.html", "/nowhere/html/bar-b.html"} == {
        p.fs_path for p in html_dir.iterdir
    }

    assert "foo-udeb" in result
    foo_udeb_fs_root = result["foo-udeb"].fs_root

    udeb_doc_dir = foo_udeb_fs_root.lookup("/usr/share/doc")
    assert udeb_doc_dir is None


def test_bts_1093680_crash_inactive_package(
    flexible_manifest_parser: FlexibleManifestParser,
) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere"),
            virtual_path_def("examples/", fs_path="/nowhere/examples"),
            virtual_path_def("examples/case-a/", fs_path="/nowhere/examples/case-a"),
            virtual_path_def(
                "examples/case-a/data/", fs_path="/nowhere/examples/case-a/data"
            ),
            virtual_path_def(
                "examples/case-a/data/foo.txt",
                fs_path="/nowhere/examples/case-a/data/foo.txt",
                content="data file",
            ),
            virtual_path_def("examples/case-b/", fs_path="/nowhere/examples/case-b"),
            virtual_path_def(
                "examples/case-b/data/", fs_path="/nowhere/examples/case-b/data"
            ),
            virtual_path_def(
                "examples/case-b/data/foo.txt",
                fs_path="/nowhere/examples/case-b/data/foo.txt",
                content="data file",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install-examples:
              source: "examples/*"
              into: foo-data
        """
    )

    foo_fields = Deb822(
        {
            "Package": "foo",
            "Architecture": "any",
        }
    )
    foo_data_fields = Deb822(
        {
            "Package": "foo-data",
            "Architecture": "all",
        }
    )
    # An arch:any build (-B => foo is active, foo-data is inactive)
    flexible_manifest_parser.binary_package(foo_fields, is_main_package=True)
    flexible_manifest_parser.binary_package(foo_data_fields, should_be_acted_on=False)

    manifest = flexible_manifest_parser.parse_manifest(manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [debian_source_root_dir],
        ),
    )


@pytest.mark.parametrize(
    "should_act_on_foo,should_act_on_foo_data",
    [
        (True, True),
        (True, False),
        (False, False),
        (False, False),
    ],
)
def test_debputy_132(
    flexible_manifest_parser: FlexibleManifestParser,
    should_act_on_foo: bool,
    should_act_on_foo_data: bool,
) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere"),
            virtual_path_def("built/", fs_path="/nowhere/built"),
            virtual_path_def("built/data/", fs_path="/nowhere/built/data"),
            virtual_path_def("built/data/foo.txt", fs_path="/nowhere/data/foo.txt"),
            virtual_path_def("built/bin/", fs_path="/nowhere/built/bin"),
            virtual_path_def(
                "built/bin/foo",
                fs_path="/nowhere/built/bin/foo",
                content="#!/bin/sh\n\nexit 0",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              source: "built/data"
              dest-dir: "usr/share/foo"
              into: foo-data
          - install:
              source: "built/bin"
              dest-dir: "usr"
              into: foo
        """
    )

    foo_fields = Deb822(
        {
            "Package": "foo",
            "Architecture": "any",
        }
    )
    foo_data_fields = Deb822(
        {
            "Package": "foo-data",
            "Architecture": "all",
        }
    )
    # An arch:any build (-B => foo is active, foo-data is inactive)
    flexible_manifest_parser.binary_package(
        foo_fields,
        is_main_package=True,
        should_be_acted_on=should_act_on_foo,
    )
    flexible_manifest_parser.binary_package(
        foo_data_fields,
        should_be_acted_on=should_act_on_foo_data,
    )

    manifest = flexible_manifest_parser.parse_manifest(manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    manifest.perform_installations(
        INTEGRATION_MODE_DH_DEBPUTY,
        install_request_context=InstallSearchDirContext(
            [
                SearchDir(debian_source_root_dir, all_pkgs),
            ],
            [debian_source_root_dir],
        ),
    )


def test_uncovered_installation(
    manifest_parser_pkg_foo: YAMLManifestParser,
) -> None:
    debian_source_root_dir = build_virtual_file_system(
        [
            virtual_path_def(".", fs_path="/nowhere"),
            virtual_path_def("usr/", fs_path="/nowhere/debian/tmp/usr"),
            virtual_path_def("usr/bin/", fs_path="/nowhere/debian/tmp/usr/bin"),
            virtual_path_def(
                "usr/bin/foo",
                fs_path="/nowhere/debian/tmp/usr/bin/foo",
                content="#!/bin/sh\n\nexit 0",
            ),
            virtual_path_def(
                "usr/bin/bar",
                fs_path="/nowhere/debian/tmp/usr/bin/bar",
                content="#!/bin/sh\n\nexit 0",
            ),
        ]
    )
    manifest_content = textwrap.dedent(
        """\
        manifest-version: "0.1"
        installations:
          - install:
              source: "usr/bin/foo"
              into: foo
        """
    )

    manifest = manifest_parser_pkg_foo.parse_manifest(fd=manifest_content)
    all_pkgs = frozenset(manifest.all_packages)

    with pytest.raises(PathNotCoveredByInstallRulesError) as exec_info:
        manifest.perform_installations(
            INTEGRATION_MODE_DH_DEBPUTY,
            install_request_context=InstallSearchDirContext(
                [
                    SearchDir(debian_source_root_dir, all_pkgs),
                ],
                [debian_source_root_dir],
            ),
        )

    assert len(exec_info.value.unmatched_paths) == 1
    assert exec_info.value.unmatched_paths[0].path == "./usr/bin/bar"
    assert exec_info.value.search_dir is debian_source_root_dir



from _typeshed import Incomplete

class InvalidColor(Exception):
    message: Incomplete
    def __init__(self, message: str) -> None: ...

class InvalidHexColor(Exception):
    message: Incomplete
    def __init__(self, message: str) -> None: ...

class InvalidStyle(Exception):
    message: Incomplete
    def __init__(self, message: str) -> None: ...

class InvalidControl(Exception):
    message: Incomplete
    def __init__(self, message: str) -> None: ...



import collections
import dataclasses
import typing
from typing import (
    Any,
    Callable,
    Tuple,
    TypedDict,
    Dict,
    get_type_hints,
    Annotated,
    get_args,
    get_origin,
    TypeVar,
    Generic,
    FrozenSet,
    Mapping,
    Optional,
    cast,
    Type,
    Union,
    List,
    Collection,
    NotRequired,
    Iterable,
    Literal,
    Sequence,
    Container,
)

from debputy.lsp.diagnostics import LintSeverity
from debputy.manifest_parser.base_types import FileSystemMatchRule
from debputy.manifest_parser.exceptions import (
    ManifestParseException,
)
from debputy.manifest_parser.mapper_code import (
    normalize_into_list,
    wrap_into_list,
    map_each_element,
)
from debputy.manifest_parser.parse_hints import (
    ConditionalRequired,
    DebputyParseHint,
    TargetAttribute,
    ManifestAttribute,
    ConflictWithSourceAttribute,
    NotPathHint,
)
from debputy.manifest_parser.parser_data import ParserContextData
from debputy.manifest_parser.tagging_types import (
    DebputyParsedContent,
    DebputyDispatchableType,
    TypeMapping,
)
from debputy.manifest_parser.util import (
    AttributePath,
    unpack_type,
    find_annotation,
    check_integration_mode,
)
from debputy.plugin.api.impl_types import (
    DeclarativeInputParser,
    TD,
    ListWrappedDeclarativeInputParser,
    DispatchingObjectParser,
    DispatchingTableParser,
    TTP,
    TP,
    InPackageContextParser,
)
from debputy.plugin.api.spec import (
    ParserDocumentation,
    DebputyIntegrationMode,
    StandardParserAttributeDocumentation,
    undocumented_attr,
    ParserAttributeDocumentation,
    reference_documentation,
)
from debputy.util import _info, _warn, assume_not_none

if typing.TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types

try:
    from Levenshtein import distance
except ImportError:
    _WARN_ONCE = False

    def _detect_possible_typo(
        _key: str,
        _value: object,
        _manifest_attributes: Mapping[str, "AttributeDescription"],
        _path: "AttributePath",
    ) -> None:
        global _WARN_ONCE
        if not _WARN_ONCE:
            _WARN_ONCE = True
            _info(
                "Install python3-levenshtein to have debputy try to detect typos in the manifest."
            )

else:

    def _detect_possible_typo(
        key: str,
        value: object,
        manifest_attributes: Mapping[str, "AttributeDescription"],
        path: "AttributePath",
    ) -> None:
        k_len = len(key)
        key_path = path[key]
        matches: List[str] = []
        current_match_strength = 0
        for acceptable_key, attr in manifest_attributes.items():
            if abs(k_len - len(acceptable_key)) > 2:
                continue
            d = distance(key, acceptable_key)
            if d > 2:
                continue
            try:
                attr.type_validator.ensure_type(value, key_path)
            except ManifestParseException:
                if attr.type_validator.base_type_match(value):
                    match_strength = 1
                else:
                    match_strength = 0
            else:
                match_strength = 2

            if match_strength < current_match_strength:
                continue
            if match_strength > current_match_strength:
                current_match_strength = match_strength
                matches.clear()
            matches.append(acceptable_key)

        if not matches:
            return
        ref = f'at "{path.path}"' if path else "at the manifest root level"
        if len(matches) == 1:
            possible_match = repr(matches[0])
            _warn(
                f'Possible typo: The key "{key}" {ref} should probably have been {possible_match}'
            )
        else:
            matches.sort()
            possible_matches = ", ".join(repr(a) for a in matches)
            _warn(
                f'Possible typo: The key "{key}" {ref} should probably have been one of {possible_matches}'
            )


SF = TypeVar("SF")
T = TypeVar("T")
S = TypeVar("S")


_NONE_TYPE = type(None)


# These must be able to appear in an "isinstance" check and must be builtin types.
BASIC_SIMPLE_TYPES = {
    str: "string",
    int: "integer",
    bool: "boolean",
}


class AttributeTypeHandler:
    __slots__ = ("_description", "_ensure_type", "base_type", "mapper")

    def __init__(
        self,
        description: str,
        ensure_type: Callable[[Any, AttributePath], None],
        *,
        base_type: Optional[Type[Any]] = None,
        mapper: Optional[
            Callable[[Any, AttributePath, Optional["ParserContextData"]], Any]
        ] = None,
    ) -> None:
        self._description = description
        self._ensure_type = ensure_type
        self.base_type = base_type
        self.mapper = mapper

    def describe_type(self) -> str:
        return self._description

    def ensure_type(self, obj: object, path: AttributePath) -> None:
        self._ensure_type(obj, path)

    def base_type_match(self, obj: object) -> bool:
        base_type = self.base_type
        return base_type is not None and isinstance(obj, base_type)

    def map_type(
        self,
        value: Any,
        path: AttributePath,
        parser_context: Optional["ParserContextData"],
    ) -> Any:
        mapper = self.mapper
        if mapper is not None:
            return mapper(value, path, parser_context)
        return value

    def combine_mapper(
        self,
        mapper: Optional[
            Callable[[Any, AttributePath, Optional["ParserContextData"]], Any]
        ],
    ) -> "AttributeTypeHandler":
        if mapper is None:
            return self
        if self.mapper is not None:
            m = self.mapper

            def _combined_mapper(
                value: Any,
                path: AttributePath,
                parser_context: Optional["ParserContextData"],
            ) -> Any:
                return mapper(m(value, path, parser_context), path, parser_context)

        else:
            _combined_mapper = mapper

        return AttributeTypeHandler(
            self._description,
            self._ensure_type,
            base_type=self.base_type,
            mapper=_combined_mapper,
        )


@dataclasses.dataclass(slots=True)
class AttributeDescription:
    source_attribute_name: str
    target_attribute: str
    attribute_type: Any
    type_validator: AttributeTypeHandler
    annotations: Tuple[Any, ...]
    conflicting_attributes: FrozenSet[str]
    conditional_required: Optional["ConditionalRequired"]
    parse_hints: Optional["DetectedDebputyParseHint"] = None
    is_optional: bool = False


def _extract_path_hint(v: Any, attribute_path: AttributePath) -> bool:
    if attribute_path.path_hint is not None:
        return True
    if isinstance(v, str):
        attribute_path.path_hint = v
        return True
    elif isinstance(v, list) and len(v) > 0 and isinstance(v[0], str):
        attribute_path.path_hint = v[0]
        return True
    return False


@dataclasses.dataclass(slots=True, frozen=True)
class DeclarativeNonMappingInputParser(DeclarativeInputParser[TD], Generic[TD, SF]):
    alt_form_parser: AttributeDescription
    inline_reference_documentation: Optional[ParserDocumentation] = None
    expected_debputy_integration_mode: Optional[Container[DebputyIntegrationMode]] = (
        None
    )

    def parse_input(
        self,
        value: object,
        path: AttributePath,
        *,
        parser_context: Optional["ParserContextData"] = None,
    ) -> TD:
        check_integration_mode(
            path,
            parser_context,
            self.expected_debputy_integration_mode,
        )
        if self.reference_documentation_url is not None:
            doc_ref = f" (Documentation: {self.reference_documentation_url})"
        else:
            doc_ref = ""

        alt_form_parser = self.alt_form_parser
        if value is None:
            form_note = f" The value must have type: {alt_form_parser.type_validator.describe_type()}"
            if self.reference_documentation_url is not None:
                doc_ref = f" Please see {self.reference_documentation_url} for the documentation."
            raise ManifestParseException(
                f"The attribute {path.path} was missing a value. {form_note}{doc_ref}"
            )
        _extract_path_hint(value, path)
        alt_form_parser.type_validator.ensure_type(value, path)
        attribute = alt_form_parser.target_attribute
        alias_mapping = {
            attribute: ("", None),
        }
        v = alt_form_parser.type_validator.map_type(value, path, parser_context)
        path.alias_mapping = alias_mapping
        return cast("TD", {attribute: v})


@dataclasses.dataclass(slots=True)
class DeclarativeMappingInputParser(DeclarativeInputParser[TD], Generic[TD, SF]):
    input_time_required_parameters: FrozenSet[str]
    all_parameters: FrozenSet[str]
    manifest_attributes: Mapping[str, "AttributeDescription"]
    source_attributes: Mapping[str, "AttributeDescription"]
    at_least_one_of: FrozenSet[FrozenSet[str]]
    alt_form_parser: Optional[AttributeDescription]
    mutually_exclusive_attributes: FrozenSet[FrozenSet[str]] = frozenset()
    _per_attribute_conflicts_cache: Optional[Mapping[str, FrozenSet[str]]] = None
    inline_reference_documentation: Optional[ParserDocumentation] = None
    path_hint_source_attributes: Sequence[str] = tuple()
    expected_debputy_integration_mode: Optional[Container[DebputyIntegrationMode]] = (
        None
    )

    def _parse_alt_form(
        self,
        value: object,
        path: AttributePath,
        *,
        parser_context: Optional["ParserContextData"] = None,
    ) -> TD:
        alt_form_parser = self.alt_form_parser
        if alt_form_parser is None:
            raise ManifestParseException(
                f"The attribute {path.path} must be a mapping.{self._doc_url_error_suffix()}"
            )
        _extract_path_hint(value, path)
        alt_form_parser.type_validator.ensure_type(value, path)
        assert (
            value is not None
        ), "The alternative form was None, but the parser should have rejected None earlier."
        attribute = alt_form_parser.target_attribute
        alias_mapping = {
            attribute: ("", None),
        }
        v = alt_form_parser.type_validator.map_type(value, path, parser_context)
        path.alias_mapping = alias_mapping
        return cast("TD", {attribute: v})

    def _validate_expected_keys(
        self,
        value: Dict[Any, Any],
        path: AttributePath,
        *,
        parser_context: Optional["ParserContextData"] = None,
    ) -> None:
        unknown_keys = value.keys() - self.all_parameters
        doc_ref = self._doc_url_error_suffix()
        if unknown_keys:
            for k in unknown_keys:
                if isinstance(k, str):
                    _detect_possible_typo(k, value[k], self.manifest_attributes, path)
            unused_keys = self.all_parameters - value.keys()
            if unused_keys:
                k = ", ".join(unused_keys)
                raise ManifestParseException(
                    f'Unknown keys "{unknown_keys}" at {path.path_container_lc}".  Keys that could be used here are: {k}.{doc_ref}'
                )
            raise ManifestParseException(
                f'Unknown keys "{unknown_keys}" at {path.path_container_lc}".  Please remove them.{doc_ref}'
            )
        missing_keys = self.input_time_required_parameters - value.keys()
        if missing_keys:
            required = ", ".join(repr(k) for k in sorted(missing_keys))
            raise ManifestParseException(
                f"The following keys were required but not present at {path.path_container_lc}: {required}{doc_ref}"
            )
        for maybe_required in self.all_parameters - value.keys():
            attr = self.manifest_attributes[maybe_required]
            assert attr.conditional_required is None or parser_context is not None
            if (
                attr.conditional_required is not None
                and attr.conditional_required.condition_applies(
                    assume_not_none(parser_context)
                )
            ):
                reason = attr.conditional_required.reason
                raise ManifestParseException(
                    f'Missing the *conditionally* required attribute "{maybe_required}" at {path.path_container_lc}. {reason}{doc_ref}'
                )
        for keyset in self.at_least_one_of:
            matched_keys = value.keys() & keyset
            if not matched_keys:
                conditionally_required = ", ".join(repr(k) for k in sorted(keyset))
                raise ManifestParseException(
                    f"At least one of the following keys must be present at {path.path_container_lc}:"
                    f" {conditionally_required}{doc_ref}"
                )
        for group in self.mutually_exclusive_attributes:
            matched = value.keys() & group
            if len(matched) > 1:
                ck = ", ".join(repr(k) for k in sorted(matched))
                raise ManifestParseException(
                    f"Could not parse {path.path_container_lc}: The following attributes are"
                    f" mutually exclusive: {ck}{doc_ref}"
                )

    def _parse_typed_dict_form(
        self,
        value: Dict[Any, Any],
        path: AttributePath,
        *,
        parser_context: Optional["ParserContextData"] = None,
    ) -> TD:
        self._validate_expected_keys(value, path, parser_context=parser_context)
        result = {}
        per_attribute_conflicts = self._per_attribute_conflicts()
        alias_mapping = {}
        for path_hint_source_attributes in self.path_hint_source_attributes:
            v = value.get(path_hint_source_attributes)
            if v is not None and _extract_path_hint(v, path):
                break
        for k, v in value.items():
            attr = self.manifest_attributes[k]
            matched = value.keys() & per_attribute_conflicts[k]
            if matched:
                ck = ", ".join(repr(k) for k in sorted(matched))
                raise ManifestParseException(
                    f'The attribute "{k}" at {path.path} cannot be used with the following'
                    f" attributes: {ck}{self._doc_url_error_suffix()}"
                )
            nk = attr.target_attribute
            key_path = path[k]
            attr.type_validator.ensure_type(v, key_path)
            if v is None:
                continue
            if k != nk:
                alias_mapping[nk] = k, None
            v = attr.type_validator.map_type(v, key_path, parser_context)
            result[nk] = v
        if alias_mapping:
            path.alias_mapping = alias_mapping
        return cast("TD", result)

    def _doc_url_error_suffix(self, *, see_url_version: bool = False) -> str:
        doc_url = self.reference_documentation_url
        if doc_url is not None:
            if see_url_version:
                return f" Please see {doc_url} for the documentation."
            return f" (Documentation: {doc_url})"
        return ""

    def parse_input(
        self,
        value: object,
        path: AttributePath,
        *,
        parser_context: Optional["ParserContextData"] = None,
    ) -> TD:
        check_integration_mode(
            path,
            parser_context,
            self.expected_debputy_integration_mode,
        )
        if value is None:
            form_note = " The attribute must be a mapping."
            if self.alt_form_parser is not None:
                form_note = (
                    " The attribute can be a mapping or a non-mapping format"
                    ' (usually, "non-mapping format" means a string or a list of strings).'
                )
            doc_ref = self._doc_url_error_suffix(see_url_version=True)
            raise ManifestParseException(
                f"The attribute {path.path} was missing a value. {form_note}{doc_ref}"
            )

        if not isinstance(value, dict):
            return self._parse_alt_form(value, path, parser_context=parser_context)
        return self._parse_typed_dict_form(value, path, parser_context=parser_context)

    def _per_attribute_conflicts(self) -> Mapping[str, FrozenSet[str]]:
        conflicts = self._per_attribute_conflicts_cache
        if conflicts is not None:
            return conflicts
        attrs = self.source_attributes
        conflicts = {
            a.source_attribute_name: frozenset(
                attrs[ca].source_attribute_name for ca in a.conflicting_attributes
            )
            for a in attrs.values()
        }
        self._per_attribute_conflicts_cache = conflicts
        return self._per_attribute_conflicts_cache


def _is_path_attribute_candidate(
    source_attribute: AttributeDescription, target_attribute: AttributeDescription
) -> bool:
    if (
        source_attribute.parse_hints
        and not source_attribute.parse_hints.applicable_as_path_hint
    ):
        return False
    target_type = target_attribute.attribute_type
    _, origin, args = unpack_type(target_type, False)
    match_type = target_type
    if origin == list:
        match_type = args[0]
    return isinstance(match_type, type) and issubclass(match_type, FileSystemMatchRule)


if typing.is_typeddict(DebputyParsedContent):
    is_typeddict = typing.is_typeddict
else:

    def is_typeddict(t: Any) -> bool:
        if typing.is_typeddict(t):
            return True
        return isinstance(t, type) and issubclass(t, DebputyParsedContent)


class ParserGenerator:
    def __init__(self) -> None:
        self._registered_types: Dict[Any, TypeMapping[Any, Any]] = {}
        self._object_parsers: Dict[str, DispatchingObjectParser] = {}
        self._table_parsers: Dict[
            Type[DebputyDispatchableType], DispatchingTableParser[Any]
        ] = {}
        self._in_package_context_parser: Dict[str, Any] = {}

    def register_mapped_type(self, mapped_type: TypeMapping[Any, Any]) -> None:
        existing = self._registered_types.get(mapped_type.target_type)
        if existing is not None:
            raise ValueError(f"The type {existing} is already registered")
        self._registered_types[mapped_type.target_type] = mapped_type

    def get_mapped_type_from_target_type(
        self,
        mapped_type: Type[T],
    ) -> Optional[TypeMapping[Any, T]]:
        return self._registered_types.get(mapped_type)

    def discard_mapped_type(self, mapped_type: Type[T]) -> None:
        del self._registered_types[mapped_type]

    def add_table_parser(self, rt: Type[DebputyDispatchableType], path: str) -> None:
        assert rt not in self._table_parsers
        self._table_parsers[rt] = DispatchingTableParser(rt, path)

    def add_object_parser(
        self,
        path: str,
        *,
        parser_documentation: Optional[ParserDocumentation] = None,
        expected_debputy_integration_mode: Optional[
            Container[DebputyIntegrationMode]
        ] = None,
        unknown_keys_diagnostic_severity: Optional[LintSeverity] = "error",
    ) -> None:
        assert path not in self._in_package_context_parser
        assert path not in self._object_parsers
        self._object_parsers[path] = DispatchingObjectParser(
            path,
            parser_documentation=parser_documentation,
            expected_debputy_integration_mode=expected_debputy_integration_mode,
            unknown_keys_diagnostic_severity=unknown_keys_diagnostic_severity,
        )

    def add_in_package_context_parser(
        self,
        path: str,
        delegate: DeclarativeInputParser[Any],
    ) -> None:
        assert path not in self._in_package_context_parser
        assert path not in self._object_parsers
        self._in_package_context_parser[path] = InPackageContextParser(path, delegate)

    @property
    def dispatchable_table_parsers(
        self,
    ) -> Mapping[Type[DebputyDispatchableType], DispatchingTableParser[Any]]:
        return self._table_parsers

    @property
    def dispatchable_object_parsers(self) -> Mapping[str, DispatchingObjectParser]:
        return self._object_parsers

    def dispatch_parser_table_for(
        self, rule_type: TTP
    ) -> Optional[DispatchingTableParser[TP]]:
        return cast(
            "Optional[DispatchingTableParser[TP]]", self._table_parsers.get(rule_type)
        )

    def generate_parser(
        self,
        parsed_content: Type[TD],
        *,
        source_content: Optional[SF] = None,
        allow_optional: bool = False,
        inline_reference_documentation: Optional[ParserDocumentation] = None,
        expected_debputy_integration_mode: Optional[
            Container[DebputyIntegrationMode]
        ] = None,
        automatic_docs: Optional[
            Mapping[Type[Any], Sequence[StandardParserAttributeDocumentation]]
        ] = None,
    ) -> DeclarativeInputParser[TD]:
        """Derive a parser from a TypedDict

        Generates a parser for a segment of the manifest (think the `install-docs` snippet) from a TypedDict
        or two that are used as a description.

        In its most simple use-case, the caller provides a TypedDict of the expected attributed along with
        their types. As an example:

          >>> class InstallDocsRule(DebputyParsedContent):
          ...     sources: List[str]
          ...     into: List[str]
          >>> pg = ParserGenerator()
          >>> simple_parser = pg.generate_parser(InstallDocsRule)

        This will create a parser that would be able to interpret something like:

        ```yaml
           install-docs:
             sources: ["docs/*"]
             into: ["my-pkg"]
        ```

        While this is sufficient for programmers, it is a bit rigid for the packager writing the manifest.  Therefore,
        you can also provide a TypedDict describing the input, enabling more flexibility:

          >>> class InstallDocsRule(DebputyParsedContent):
          ...     sources: List[str]
          ...     into: List[str]
          >>> class InputDocsRuleInputFormat(TypedDict):
          ...     source: NotRequired[Annotated[str, DebputyParseHint.target_attribute("sources")]]
          ...     sources: NotRequired[List[str]]
          ...     into: Union[str, List[str]]
          >>> pg = ParserGenerator()
          >>> flexible_parser = pg.generate_parser(
          ...    InstallDocsRule,
          ...    source_content=InputDocsRuleInputFormat,
          ... )

        In this case, the `sources` field can either come from a single `source` in the manifest (which must be a string)
        or `sources` (which must be a list of strings). The parser also ensures that only one of `source` or `sources`
        is used to ensure the input is not ambiguous. For the `into` parameter, the parser will accept it being a str
        or a list of strings.  Regardless of how the input was provided, the parser will normalize the input so that
        both `sources` and `into` in the result is a list of strings.  As an example, this parser can accept
        both the previous input but also the following input:

        ```yaml
           install-docs:
             source: "docs/*"
             into: "my-pkg"
        ```

        The `source` and `into` attributes are then normalized to lists as if the user had written them as lists
        with a single string in them. As noted above, the name of the `source` attribute will also be normalized
        while parsing.

        In the cases where only one field is required by the user, it can sometimes make sense to allow a non-dict
        as part of the input.  Example:

          >>> class DiscardRule(DebputyParsedContent):
          ...     paths: List[str]
          >>> class DiscardRuleInputDictFormat(TypedDict):
          ...     path: NotRequired[Annotated[str, DebputyParseHint.target_attribute("paths")]]
          ...     paths: NotRequired[List[str]]
          >>> # This format relies on DiscardRule having exactly one Required attribute
          >>> DiscardRuleInputWithAltFormat = Union[
          ...    DiscardRuleInputDictFormat,
          ...    str,
          ...    List[str],
          ... ]
          >>> pg = ParserGenerator()
          >>> flexible_parser = pg.generate_parser(
          ...    DiscardRule,
          ...    source_content=DiscardRuleInputWithAltFormat,
          ... )


        Supported types:
          * `List` - must have a fixed type argument (such as `List[str]`)
          * `str`
          * `int`
          * `BinaryPackage` - When provided (or required), the user must provide a package name listed
                              in the debian/control file. The code receives the BinaryPackage instance
                              matching that input.
          * `FileSystemMode` - When provided (or required), the user must provide a file system mode in any
                               format that `debputy' provides (such as `0644` or `a=rw,go=rw`).
          * `FileSystemOwner` - When provided (or required), the user must a file system owner that is
                                available statically on all Debian systems (must be in `base-passwd`).
                                The user has multiple options for how to specify it (either via name or id).
          * `FileSystemGroup` - When provided (or required), the user must a file system group that is
                                available statically on all Debian systems (must be in `base-passwd`).
                                The user has multiple options for how to specify it (either via name or id).
          * `ManifestCondition` - When provided (or required), the user must specify a conditional rule to apply.
                                  Usually, it is better to extend `DebputyParsedContentStandardConditional`, which
                                  provides the `debputy' default `when` parameter for conditionals.

        Supported special type-like parameters:

          * `Required` / `NotRequired` to mark a field as `Required` or `NotRequired`. Must be provided at the
             outermost level.  Cannot vary between `parsed_content` and `source_content`.
          * `Annotated`. Accepted at the outermost level (inside Required/NotRequired) but ignored at the moment.
          * `Union`. Must be the outermost level (inside `Annotated` or/and `Required`/`NotRequired` if these are present).
            Automapping (see below) is restricted to two members in the Union.

        Notable non-supported types:
          * `Mapping` and all variants therefore (such as `dict`). In the future, nested `TypedDict`s may be allowed.
          * `Optional` (or `Union[..., None]`): Use `NotRequired` for optional fields.

        Automatic mapping rules from `source_content` to `parsed_content`:
          - `Union[T, List[T]]` can be narrowed automatically to `List[T]`.  Transformation is basically:
            `lambda value: value if isinstance(value, list) else [value]`
          - `T` can be mapped automatically to `List[T]`, Transformation being: `lambda value: [value]`

        Additionally, types can be annotated (`Annotated[str, ...]`) with `DebputyParseHint`s.  Check its classmethod
        for concrete features that may be useful to you.

        :param parsed_content: A DebputyParsedContent / TypedDict describing the desired model of the input once parsed.
          (DebputyParsedContent is a TypedDict subclass that work around some inadequate type checkers).
          It can also be a `List[DebputyParsedContent]`. In that case, `source_content` must be a
          `List[TypedDict[...]]`.
        :param source_content: Optionally, a TypedDict describing the input allowed by the user.  This can be useful
          to describe more variations than in `parsed_content` that the parser will normalize for you. If omitted,
          the parsed_content is also considered the source_content (which affects what annotations are allowed in it).
          Note you should never pass the parsed_content as source_content directly.
        :param allow_optional: In rare cases, you want to support explicitly provided vs. optional.  In this case, you
          should set this to True.  Though, in 99.9% of all cases, you want `NotRequired` rather than `Optional` (and
          can keep this False).
        :param inline_reference_documentation: Optionally, programmatic documentation
        :param expected_debputy_integration_mode: If provided, this declares the integration modes where the
          result of the parser can be used. This is primarily useful for "fail-fast" on incorrect usage.
          When the restriction is not satisfiable, the generated parser will trigger a parse error immediately
          (resulting in a "compile time" failure rather than a "runtime" failure).
        :return: An input parser capable of reading input matching the TypedDict(s) used as reference.
        """
        orig_parsed_content = parsed_content
        if source_content is parsed_content:
            raise ValueError(
                "Do not provide source_content if it is the same as parsed_content"
            )
        is_list_wrapped = False
        if get_origin(orig_parsed_content) == list:
            parsed_content = get_args(orig_parsed_content)[0]
            is_list_wrapped = True

        if isinstance(parsed_content, type) and issubclass(
            parsed_content, DebputyDispatchableType
        ):
            parser = self.dispatch_parser_table_for(parsed_content)
            if parser is None:
                raise ValueError(
                    f"Unsupported parsed_content descriptor: {parsed_content.__qualname__}."
                    f" The class {parsed_content.__qualname__} is not a pre-registered type."
                )
            # FIXME: Only the list wrapped version has documentation.
            if is_list_wrapped:
                parser = ListWrappedDeclarativeInputParser(
                    parser,
                    inline_reference_documentation=inline_reference_documentation,
                    expected_debputy_integration_mode=expected_debputy_integration_mode,
                )
            return parser

        if not is_typeddict(parsed_content):
            raise ValueError(
                f"Unsupported parsed_content descriptor: {parsed_content.__qualname__}."
                ' Only "TypedDict"-based types and a subset of "DebputyDispatchableType" are supported.'
            )
        if is_list_wrapped and source_content is not None:
            if get_origin(source_content) != list:
                raise ValueError(
                    "If the parsed_content is a List type, then source_format must be a List type as well."
                )
            source_content = get_args(source_content)[0]

        target_attributes = self._parse_types(
            parsed_content,
            allow_source_attribute_annotations=source_content is None,
            forbid_optional=not allow_optional,
        )
        required_target_parameters = frozenset(parsed_content.__required_keys__)
        parsed_alt_form = None
        non_mapping_source_only = False

        if source_content is not None:
            default_target_attribute = None
            if len(required_target_parameters) == 1:
                default_target_attribute = next(iter(required_target_parameters))

            source_typed_dict, alt_source_forms = _extract_typed_dict(
                source_content,
                default_target_attribute,
            )
            if alt_source_forms:
                parsed_alt_form = self._parse_alt_form(
                    alt_source_forms,
                    default_target_attribute,
                )
            if source_typed_dict is not None:
                source_content_attributes = self._parse_types(
                    source_typed_dict,
                    allow_target_attribute_annotation=True,
                    allow_source_attribute_annotations=True,
                    forbid_optional=not allow_optional,
                )
                source_content_parameter = "source_content"
                source_and_parsed_differs = True
            else:
                source_typed_dict = parsed_content
                source_content_attributes = target_attributes
                source_content_parameter = "parsed_content"
                source_and_parsed_differs = True
                non_mapping_source_only = True
        else:
            source_typed_dict = parsed_content
            source_content_attributes = target_attributes
            source_content_parameter = "parsed_content"
            source_and_parsed_differs = False

        sources = collections.defaultdict(set)
        seen_targets = set()
        seen_source_names: Dict[str, str] = {}
        source_attributes: Dict[str, AttributeDescription] = {}
        path_hint_source_attributes = []

        for k in source_content_attributes:
            ia = source_content_attributes[k]

            ta = (
                target_attributes.get(ia.target_attribute)
                if source_and_parsed_differs
                else ia
            )
            if ta is None:
                # Error message would be wrong if this assertion is false.
                assert source_and_parsed_differs
                raise ValueError(
                    f'The attribute "{k}" from the "source_content" parameter should have mapped'
                    f' to "{ia.target_attribute}", but that parameter does not exist in "parsed_content"'
                )
            if _is_path_attribute_candidate(ia, ta):
                path_hint_source_attributes.append(ia.source_attribute_name)
            existing_source_name = seen_source_names.get(ia.source_attribute_name)
            if existing_source_name:
                raise ValueError(
                    f'The attribute "{k}" and "{existing_source_name}" both share the source name'
                    f' "{ia.source_attribute_name}". Please change the {source_content_parameter} parameter,'
                    f' so only one attribute use "{ia.source_attribute_name}".'
                )
            seen_source_names[ia.source_attribute_name] = k
            seen_targets.add(ta.target_attribute)
            sources[ia.target_attribute].add(k)
            if source_and_parsed_differs:
                bridge_mapper = self._type_normalize(
                    k, ia.attribute_type, ta.attribute_type, False
                )
                ia.type_validator = ia.type_validator.combine_mapper(bridge_mapper)
            source_attributes[k] = ia

        def _as_attr_names(td_name: Iterable[str]) -> FrozenSet[str]:
            return frozenset(
                source_content_attributes[a].source_attribute_name for a in td_name
            )

        _check_attributes(
            parsed_content,
            source_typed_dict,
            source_content_attributes,
            sources,
        )

        at_least_one_of = frozenset(
            _as_attr_names(g)
            for k, g in sources.items()
            if len(g) > 1 and k in required_target_parameters
        )

        if source_and_parsed_differs and seen_targets != target_attributes.keys():
            missing = ", ".join(
                repr(k) for k in (target_attributes.keys() - seen_targets)
            )
            raise ValueError(
                'The following attributes in "parsed_content" did not have a source field in "source_content":'
                f" {missing}"
            )
        all_mutually_exclusive_fields = frozenset(
            _as_attr_names(g) for g in sources.values() if len(g) > 1
        )

        all_parameters = (
            source_typed_dict.__required_keys__ | source_typed_dict.__optional_keys__
        )
        _check_conflicts(
            source_content_attributes,
            source_typed_dict.__required_keys__,
            all_parameters,
        )

        manifest_attributes = {
            a.source_attribute_name: a for a in source_content_attributes.values()
        }

        if parsed_alt_form is not None:
            target_attribute = parsed_alt_form.target_attribute
            if (
                target_attribute not in required_target_parameters
                and required_target_parameters
                or len(required_target_parameters) > 1
            ):
                raise NotImplementedError(
                    "When using alternative source formats (Union[TypedDict, ...]), then the"
                    " target must have at most one require parameter"
                )
            bridge_mapper = self._type_normalize(
                target_attribute,
                parsed_alt_form.attribute_type,
                target_attributes[target_attribute].attribute_type,
                False,
            )
            parsed_alt_form.type_validator = (
                parsed_alt_form.type_validator.combine_mapper(bridge_mapper)
            )

        inline_reference_documentation = (
            _verify_and_auto_correct_inline_reference_documentation(
                parsed_content,
                source_typed_dict,
                source_content_attributes,
                inline_reference_documentation,
                parsed_alt_form is not None,
                automatic_docs,
            )
        )
        if non_mapping_source_only:
            parser = DeclarativeNonMappingInputParser(
                assume_not_none(parsed_alt_form),
                inline_reference_documentation=inline_reference_documentation,
                expected_debputy_integration_mode=expected_debputy_integration_mode,
            )
        else:
            parser = DeclarativeMappingInputParser(
                _as_attr_names(source_typed_dict.__required_keys__),
                _as_attr_names(all_parameters),
                manifest_attributes,
                source_attributes,
                mutually_exclusive_attributes=all_mutually_exclusive_fields,
                alt_form_parser=parsed_alt_form,
                at_least_one_of=at_least_one_of,
                inline_reference_documentation=inline_reference_documentation,
                path_hint_source_attributes=tuple(path_hint_source_attributes),
                expected_debputy_integration_mode=expected_debputy_integration_mode,
            )
        if is_list_wrapped:
            parser = ListWrappedDeclarativeInputParser(
                parser,
                expected_debputy_integration_mode=expected_debputy_integration_mode,
            )
        return parser

    def _as_type_validator(
        self,
        attribute: str,
        provided_type: Any,
        parsing_typed_dict_attribute: bool,
    ) -> AttributeTypeHandler:
        assert not isinstance(provided_type, tuple)

        if isinstance(provided_type, type) and issubclass(
            provided_type, DebputyDispatchableType
        ):
            return _dispatch_parser(provided_type)

        unmapped_type = self._strip_mapped_types(
            provided_type,
            parsing_typed_dict_attribute,
        )
        type_normalizer = self._type_normalize(
            attribute,
            unmapped_type,
            provided_type,
            parsing_typed_dict_attribute,
        )
        t_unmapped, t_orig, t_args = unpack_type(
            unmapped_type,
            parsing_typed_dict_attribute,
        )

        if (
            t_orig == Union
            and t_args
            and len(t_args) == 2
            and any(v is _NONE_TYPE for v in t_args)
        ):
            _, _, args = unpack_type(provided_type, parsing_typed_dict_attribute)
            actual_type = [a for a in args if a is not _NONE_TYPE][0]
            validator = self._as_type_validator(
                attribute, actual_type, parsing_typed_dict_attribute
            )

            def _validator(v: Any, path: AttributePath) -> None:
                if v is None:
                    return
                validator.ensure_type(v, path)

            return AttributeTypeHandler(
                validator.describe_type(),
                _validator,
                base_type=validator.base_type,
                mapper=type_normalizer,
            )

        if unmapped_type in BASIC_SIMPLE_TYPES:
            type_name = BASIC_SIMPLE_TYPES[unmapped_type]

            type_mapping = self._registered_types.get(provided_type)
            if type_mapping is not None:
                simple_type = f" ({type_name})"
                type_name = type_mapping.target_type.__name__
            else:
                simple_type = ""

            def _validator(v: Any, path: AttributePath) -> None:
                if not isinstance(v, unmapped_type):
                    _validation_type_error(
                        path, f"The attribute must be a {type_name}{simple_type}"
                    )

            return AttributeTypeHandler(
                type_name,
                _validator,
                base_type=unmapped_type,
                mapper=type_normalizer,
            )
        if t_orig == list:
            if not t_args:
                raise ValueError(
                    f'The attribute "{attribute}" is List but does not have Generics (Must use List[X])'
                )
            _, t_provided_orig, t_provided_args = unpack_type(
                provided_type,
                parsing_typed_dict_attribute,
            )
            genetic_type = t_args[0]
            key_mapper = self._as_type_validator(
                attribute,
                genetic_type,
                parsing_typed_dict_attribute,
            )

            def _validator(v: Any, path: AttributePath) -> None:
                if not isinstance(v, list):
                    _validation_type_error(path, "The attribute must be a list")
                for i, v in enumerate(v):
                    key_mapper.ensure_type(v, path[i])

            list_mapper = (
                map_each_element(key_mapper.mapper)
                if key_mapper.mapper is not None
                else None
            )

            return AttributeTypeHandler(
                f"List of {key_mapper.describe_type()}",
                _validator,
                base_type=list,
                mapper=type_normalizer,
            ).combine_mapper(list_mapper)
        if is_typeddict(provided_type):
            subparser = self.generate_parser(cast("Type[TD]", provided_type))
            return AttributeTypeHandler(
                description=f"{provided_type.__name__} (Typed Mapping)",
                ensure_type=lambda v, ap: None,
                base_type=dict,
                mapper=lambda v, ap, cv: subparser.parse_input(
                    v, ap, parser_context=cv
                ),
            )
        if t_orig == dict:
            if not t_args or len(t_args) != 2:
                raise ValueError(
                    f'The attribute "{attribute}" is Dict but does not have Generics (Must use Dict[str, Y])'
                )
            if t_args[0] != str:
                raise ValueError(
                    f'The attribute "{attribute}" is Dict and has a non-str type as key.'
                    " Currently, only `str` is supported (Dict[str, Y])"
                )
            key_mapper = self._as_type_validator(
                attribute,
                t_args[0],
                parsing_typed_dict_attribute,
            )
            value_mapper = self._as_type_validator(
                attribute,
                t_args[1],
                parsing_typed_dict_attribute,
            )

            if key_mapper.base_type is None:
                raise ValueError(
                    f'The attribute "{attribute}" is Dict and the key did not have a trivial base type.  Key types'
                    f" without trivial base types (such as `str`) are not supported at the moment."
                )

            if value_mapper.mapper is not None:
                raise ValueError(
                    f'The attribute "{attribute}" is Dict and the value requires mapping.'
                    " Currently, this is not supported. Consider a simpler type (such as Dict[str, str] or Dict[str, Any])."
                    " Better typing may come later"
                )

            def _validator(uv: Any, path: AttributePath) -> None:
                if not isinstance(uv, dict):
                    _validation_type_error(path, "The attribute must be a mapping")
                key_name = "the first key in the mapping"
                for i, (k, v) in enumerate(uv.items()):
                    if not key_mapper.base_type_match(k):
                        kp = path.copy_with_path_hint(key_name)
                        _validation_type_error(
                            kp,
                            f'The key number {i + 1} in attribute "{kp}" must be a {key_mapper.describe_type()}',
                        )
                    key_name = f"the key after {k}"
                    value_mapper.ensure_type(v, path[k])

            return AttributeTypeHandler(
                f"Mapping of {value_mapper.describe_type()}",
                _validator,
                base_type=dict,
                mapper=type_normalizer,
            ).combine_mapper(key_mapper.mapper)
        if t_orig == Union:
            if _is_two_arg_x_list_x(t_args):
                # Force the order to be "X, List[X]" as it simplifies the code
                x_list_x = (
                    t_args if get_origin(t_args[1]) == list else (t_args[1], t_args[0])
                )

                # X, List[X] could match if X was List[Y].  However, our code below assumes
                # that X is a non-list.  The `_is_two_arg_x_list_x` returns False for this
                # case to avoid this assert and fall into the "generic case".
                assert get_origin(x_list_x[0]) != list
                x_subtype_checker = self._as_type_validator(
                    attribute,
                    x_list_x[0],
                    parsing_typed_dict_attribute,
                )
                list_x_subtype_checker = self._as_type_validator(
                    attribute,
                    x_list_x[1],
                    parsing_typed_dict_attribute,
                )
                type_description = x_subtype_checker.describe_type()
                type_description = f"{type_description} or a list of {type_description}"

                def _validator(v: Any, path: AttributePath) -> None:
                    if isinstance(v, list):
                        list_x_subtype_checker.ensure_type(v, path)
                    else:
                        x_subtype_checker.ensure_type(v, path)

                return AttributeTypeHandler(
                    type_description,
                    _validator,
                    mapper=type_normalizer,
                )
            else:
                subtype_checker = [
                    self._as_type_validator(attribute, a, parsing_typed_dict_attribute)
                    for a in t_args
                ]
                type_description = "one-of: " + ", ".join(
                    f"{sc.describe_type()}" for sc in subtype_checker
                )
                mapper = subtype_checker[0].mapper
                if any(mapper != sc.mapper for sc in subtype_checker):
                    raise ValueError(
                        f'Cannot handle the union "{provided_type}" as the target types need different'
                        " type normalization/mapping logic.  Unions are generally limited to Union[X, List[X]]"
                        " where X is a non-collection type."
                    )

                def _validator(v: Any, path: AttributePath) -> None:
                    partial_matches = []
                    for sc in subtype_checker:
                        try:
                            sc.ensure_type(v, path)
                            return
                        except ManifestParseException as e:
                            if sc.base_type_match(v):
                                partial_matches.append((sc, e))

                    if len(partial_matches) == 1:
                        raise partial_matches[0][1]
                    _validation_type_error(
                        path, f"Could not match against: {type_description}"
                    )

                return AttributeTypeHandler(
                    type_description,
                    _validator,
                    mapper=type_normalizer,
                )
        if t_orig == Literal:
            # We want "x" for string values; repr provides 'x'
            pretty = ", ".join(
                f'"{v}"' if isinstance(v, str) else str(v) for v in t_args
            )

            def _validator(v: Any, path: AttributePath) -> None:
                if v not in t_args:
                    value_hint = ""
                    if isinstance(v, str):
                        value_hint = f"({v}) "
                    _validation_type_error(
                        path,
                        f"Value {value_hint}must be one of the following literal values: {pretty}",
                    )

            return AttributeTypeHandler(
                f"One of the following literal values: {pretty}",
                _validator,
            )

        if provided_type == Any:
            return AttributeTypeHandler(
                "any (unvalidated)",
                lambda *a: None,
            )
        raise ValueError(
            f'The attribute "{attribute}" had/contained a type {provided_type}, which is not supported'
        )

    def _parse_types(
        self,
        spec: Type[TypedDict],
        allow_target_attribute_annotation: bool = False,
        allow_source_attribute_annotations: bool = False,
        forbid_optional: bool = True,
    ) -> Dict[str, AttributeDescription]:
        annotations = get_type_hints(spec, include_extras=True)
        return {
            k: self._attribute_description(
                k,
                t,
                k in spec.__required_keys__,
                allow_target_attribute_annotation=allow_target_attribute_annotation,
                allow_source_attribute_annotations=allow_source_attribute_annotations,
                forbid_optional=forbid_optional,
            )
            for k, t in annotations.items()
        }

    def _attribute_description(
        self,
        attribute: str,
        orig_td: Any,
        is_required: bool,
        forbid_optional: bool = True,
        allow_target_attribute_annotation: bool = False,
        allow_source_attribute_annotations: bool = False,
    ) -> AttributeDescription:
        td, anno, is_optional = _parse_type(
            attribute, orig_td, forbid_optional=forbid_optional
        )
        type_validator = self._as_type_validator(attribute, td, True)
        parsed_annotations = DetectedDebputyParseHint.parse_annotations(
            anno,
            f' Seen with attribute "{attribute}".',
            attribute,
            is_required,
            allow_target_attribute_annotation=allow_target_attribute_annotation,
            allow_source_attribute_annotations=allow_source_attribute_annotations,
        )
        return AttributeDescription(
            target_attribute=parsed_annotations.target_attribute,
            attribute_type=td,
            type_validator=type_validator,
            annotations=anno,
            is_optional=is_optional,
            conflicting_attributes=parsed_annotations.conflict_with_source_attributes,
            conditional_required=parsed_annotations.conditional_required,
            source_attribute_name=assume_not_none(
                parsed_annotations.source_manifest_attribute
            ),
            parse_hints=parsed_annotations,
        )

    def _parse_alt_form(
        self,
        alt_form,
        default_target_attribute: Optional[str],
    ) -> AttributeDescription:
        td, anno, is_optional = _parse_type(
            "source_format alternative form",
            alt_form,
            forbid_optional=True,
            parsing_typed_dict_attribute=False,
        )
        type_validator = self._as_type_validator(
            "source_format alternative form",
            td,
            True,
        )
        parsed_annotations = DetectedDebputyParseHint.parse_annotations(
            anno,
            " The alternative for source_format.",
            None,
            False,
            default_target_attribute=default_target_attribute,
            allow_target_attribute_annotation=True,
            allow_source_attribute_annotations=False,
        )
        return AttributeDescription(
            target_attribute=parsed_annotations.target_attribute,
            attribute_type=td,
            type_validator=type_validator,
            annotations=anno,
            is_optional=is_optional,
            conflicting_attributes=parsed_annotations.conflict_with_source_attributes,
            conditional_required=parsed_annotations.conditional_required,
            source_attribute_name="Alt form of the source_format",
        )

    def _union_narrowing(
        self,
        input_type: Any,
        target_type: Any,
        parsing_typed_dict_attribute: bool,
    ) -> Optional[Callable[[Any, AttributePath, Optional["ParserContextData"]], Any]]:
        _, input_orig, input_args = unpack_type(
            input_type, parsing_typed_dict_attribute
        )
        _, target_orig, target_args = unpack_type(
            target_type, parsing_typed_dict_attribute
        )

        if input_orig != Union or not input_args:
            raise ValueError("input_type must be a Union[...] with non-empty args")

        # Currently, we only support Union[X, List[X]] -> List[Y] narrowing or Union[X, List[X]] -> Union[Y, Union[Y]]
        # - Where X = Y or there is a simple standard transformation from X to Y.

        if target_orig not in (Union, list) or not target_args:
            # Not supported
            return None

        if target_orig == Union and set(input_args) == set(target_args):
            # Not needed (identity mapping)
            return None

        if target_orig == list and not any(get_origin(a) == list for a in input_args):
            # Not supported
            return None

        target_arg = target_args[0]
        simplified_type = self._strip_mapped_types(
            target_arg, parsing_typed_dict_attribute
        )
        acceptable_types = {
            target_arg,
            List[target_arg],  # type: ignore
            simplified_type,
            List[simplified_type],  # type: ignore
        }
        target_format = (
            target_arg,
            List[target_arg],  # type: ignore
        )
        in_target_format = 0
        in_simple_format = 0
        for input_arg in input_args:
            if input_arg not in acceptable_types:
                # Not supported
                return None
            if input_arg in target_format:
                in_target_format += 1
            else:
                in_simple_format += 1

        assert in_simple_format or in_target_format

        if in_target_format and not in_simple_format:
            # Union[X, List[X]] -> List[X]
            return normalize_into_list
        mapped = self._registered_types[target_arg]
        if not in_target_format and in_simple_format:
            # Union[X, List[X]] -> List[Y]

            def _mapper_x_list_y(
                x: Union[Any, List[Any]],
                ap: AttributePath,
                pc: Optional["ParserContextData"],
            ) -> List[Any]:
                in_list_form: List[Any] = normalize_into_list(x, ap, pc)

                return [mapped.mapper(x, ap, pc) for x in in_list_form]

            return _mapper_x_list_y

        # Union[Y, List[X]] -> List[Y]
        if not isinstance(target_arg, type):
            raise ValueError(
                f"Cannot narrow {input_type} -> {target_type}: The automatic conversion does"
                f" not support mixed types.  Please use either {simplified_type} or {target_arg}"
                f" in the source content (but both a mix of both)"
            )

        def _mapper_mixed_list_y(
            x: Union[Any, List[Any]],
            ap: AttributePath,
            pc: Optional["ParserContextData"],
        ) -> List[Any]:
            in_list_form: List[Any] = normalize_into_list(x, ap, pc)

            return [
                x if isinstance(x, target_arg) else mapped.mapper(x, ap, pc)
                for x in in_list_form
            ]

        return _mapper_mixed_list_y

    def _type_normalize(
        self,
        attribute: str,
        input_type: Any,
        target_type: Any,
        parsing_typed_dict_attribute: bool,
    ) -> Optional[Callable[[Any, AttributePath, Optional["ParserContextData"]], Any]]:
        if input_type == target_type:
            return None
        _, input_orig, input_args = unpack_type(
            input_type, parsing_typed_dict_attribute
        )
        _, target_orig, target_args = unpack_type(
            target_type,
            parsing_typed_dict_attribute,
        )
        if input_orig == Union:
            result = self._union_narrowing(
                input_type, target_type, parsing_typed_dict_attribute
            )
            if result:
                return result
        elif target_orig == list and target_args[0] == input_type:
            return wrap_into_list

        mapped = self._registered_types.get(target_type)
        if mapped is not None and input_type == mapped.source_type:
            # Source -> Target
            return mapped.mapper
        if target_orig == list and target_args:
            mapped = self._registered_types.get(target_args[0])
            if mapped is not None:
                # mypy is dense and forgot `mapped` cannot be optional in the comprehensions.
                mapped_type: TypeMapping = mapped
                if input_type == mapped.source_type:
                    # Source -> List[Target]
                    return lambda x, ap, pc: [mapped_type.mapper(x, ap, pc)]
                if (
                    input_orig == list
                    and input_args
                    and input_args[0] == mapped_type.source_type
                ):
                    # List[Source] -> List[Target]
                    return lambda xs, ap, pc: [
                        mapped_type.mapper(x, ap, pc) for x in xs
                    ]

        raise ValueError(
            f'Unsupported type normalization for "{attribute}": Cannot automatically map/narrow'
            f" {input_type} to {target_type}"
        )

    def _strip_mapped_types(
        self, orig_td: Any, parsing_typed_dict_attribute: bool
    ) -> Any:
        m = self._registered_types.get(orig_td)
        if m is not None:
            return m.source_type
        _, v, args = unpack_type(orig_td, parsing_typed_dict_attribute)
        if v == list:
            arg = args[0]
            m = self._registered_types.get(arg)
            if m:
                return List[m.source_type]  # type: ignore
        if v == Union:
            stripped_args = tuple(
                self._strip_mapped_types(x, parsing_typed_dict_attribute) for x in args
            )
            if stripped_args != args:
                return Union[stripped_args]
        return orig_td


def _sort_key(attr: StandardParserAttributeDocumentation) -> Any:
    key = next(iter(attr.attributes))
    return attr.sort_category, key


def _apply_std_docs(
    std_doc_table: Optional[
        Mapping[Type[Any], Sequence[StandardParserAttributeDocumentation]]
    ],
    source_format_typed_dict: Type[Any],
    attribute_docs: Optional[Sequence[ParserAttributeDocumentation]],
) -> Optional[Sequence[ParserAttributeDocumentation]]:
    if std_doc_table is None or not std_doc_table:
        return attribute_docs

    has_docs_for = set()
    if attribute_docs:
        for attribute_doc in attribute_docs:
            has_docs_for.update(attribute_doc.attributes)

    base_seen = set()
    std_docs_used = []

    remaining_bases = set(getattr(source_format_typed_dict, "__orig_bases__", []))
    base_seen.update(remaining_bases)
    while remaining_bases:
        base = remaining_bases.pop()
        new_bases_to_check = {
            x for x in getattr(base, "__orig_bases__", []) if x not in base_seen
        }
        remaining_bases.update(new_bases_to_check)
        base_seen.update(new_bases_to_check)
        std_docs = std_doc_table.get(base)
        if std_docs:
            for std_doc in std_docs:
                if any(a in has_docs_for for a in std_doc.attributes):
                    # If there is any overlap, do not add the docs
                    continue
                has_docs_for.update(std_doc.attributes)
                std_docs_used.append(std_doc)

    if not std_docs_used:
        return attribute_docs
    docs = sorted(std_docs_used, key=_sort_key)
    if attribute_docs:
        # Plugin provided attributes first
        c = list(attribute_docs)
        c.extend(docs)
        docs = c
    return tuple(docs)


def _verify_and_auto_correct_inline_reference_documentation(
    parsed_content: Type[TD],
    source_typed_dict: Type[Any],
    source_content_attributes: Mapping[str, AttributeDescription],
    inline_reference_documentation: Optional[ParserDocumentation],
    has_alt_form: bool,
    automatic_docs: Optional[
        Mapping[Type[Any], Sequence[StandardParserAttributeDocumentation]]
    ] = None,
) -> Optional[ParserDocumentation]:
    orig_attribute_docs = (
        inline_reference_documentation.attribute_doc
        if inline_reference_documentation
        else None
    )
    attribute_docs = _apply_std_docs(
        automatic_docs,
        source_typed_dict,
        orig_attribute_docs,
    )
    if inline_reference_documentation is None and attribute_docs is None:
        return None
    changes = {}
    if attribute_docs:
        seen = set()
        had_any_custom_docs = False
        for attr_doc in attribute_docs:
            if not isinstance(attr_doc, StandardParserAttributeDocumentation):
                had_any_custom_docs = True
            for attr_name in attr_doc.attributes:
                attr = source_content_attributes.get(attr_name)
                if attr is None:
                    raise ValueError(
                        f"The inline_reference_documentation for the source format of {parsed_content.__qualname__}"
                        f' references an attribute "{attr_name}", which does not exist in the source format.'
                    )
                if attr_name in seen:
                    raise ValueError(
                        f"The inline_reference_documentation for the source format of {parsed_content.__qualname__}"
                        f' has documentation for "{attr_name}" twice, which is not supported.'
                        f" Please document it at most once"
                    )
                seen.add(attr_name)
        undocumented = source_content_attributes.keys() - seen
        if undocumented:
            if had_any_custom_docs:
                undocumented_attrs = ", ".join(undocumented)
                raise ValueError(
                    f"The following attributes were not documented for the source format of"
                    f" {parsed_content.__qualname__}.  If this is deliberate, then please"
                    ' declare each them as undocumented (via undocumented_attr("foo")):'
                    f" {undocumented_attrs}"
                )
            combined_docs = list(attribute_docs)
            combined_docs.extend(undocumented_attr(a) for a in sorted(undocumented))
            attribute_docs = combined_docs

    if attribute_docs and orig_attribute_docs != attribute_docs:
        assert attribute_docs is not None
        changes["attribute_doc"] = tuple(attribute_docs)

    if (
        inline_reference_documentation is not None
        and inline_reference_documentation.alt_parser_description
        and not has_alt_form
    ):
        raise ValueError(
            "The inline_reference_documentation had documentation for an non-mapping format,"
            " but the source format does not have a non-mapping format."
        )
    if changes:
        if inline_reference_documentation is None:
            inline_reference_documentation = reference_documentation()
        return inline_reference_documentation.replace(**changes)
    return inline_reference_documentation


def _check_conflicts(
    input_content_attributes: Dict[str, AttributeDescription],
    required_attributes: FrozenSet[str],
    all_attributes: FrozenSet[str],
) -> None:
    for attr_name, attr in input_content_attributes.items():
        if attr_name in required_attributes and attr.conflicting_attributes:
            c = ", ".join(repr(a) for a in attr.conflicting_attributes)
            raise ValueError(
                f'The attribute "{attr_name}" is required and conflicts with the attributes: {c}.'
                " This makes it impossible to use these attributes. Either remove the attributes"
                f' (along with the conflicts for them), adjust the conflicts or make "{attr_name}"'
                " optional (NotRequired)"
            )
        else:
            required_conflicts = attr.conflicting_attributes & required_attributes
            if required_conflicts:
                c = ", ".join(repr(a) for a in required_conflicts)
                raise ValueError(
                    f'The attribute "{attr_name}" conflicts with the following *required* attributes: {c}.'
                    f' This makes it impossible to use the "{attr_name}" attribute. Either remove it,'
                    f" adjust the conflicts or make the listed attributes optional (NotRequired)"
                )
        unknown_attributes = attr.conflicting_attributes - all_attributes
        if unknown_attributes:
            c = ", ".join(repr(a) for a in unknown_attributes)
            raise ValueError(
                f'The attribute "{attr_name}" declares a conflict with the following unknown attributes: {c}.'
                f" None of these attributes were declared in the input."
            )


def _check_attributes(
    content: Type[TypedDict],
    input_content: Type[TypedDict],
    input_content_attributes: Dict[str, AttributeDescription],
    sources: Mapping[str, Collection[str]],
) -> None:
    target_required_keys = content.__required_keys__
    input_required_keys = input_content.__required_keys__
    all_input_keys = input_required_keys | input_content.__optional_keys__

    for input_name in all_input_keys:
        attr = input_content_attributes[input_name]
        target_name = attr.target_attribute
        source_names = sources[target_name]
        input_is_required = input_name in input_required_keys
        target_is_required = target_name in target_required_keys

        assert source_names

        if input_is_required and len(source_names) > 1:
            raise ValueError(
                f'The source attribute "{input_name}" is required, but it maps to "{target_name}",'
                f' which has multiple sources "{source_names}". If "{input_name}" should be required,'
                f' then there is no need for additional sources for "{target_name}". Alternatively,'
                f' "{input_name}" might be missing a NotRequired type'
                f' (example: "{input_name}: NotRequired[<OriginalTypeHere>]")'
            )
        if not input_is_required and target_is_required and len(source_names) == 1:
            raise ValueError(
                f'The source attribute "{input_name}" is not marked as required and maps to'
                f' "{target_name}", which is marked as required. As there are no other attributes'
                f' mapping to "{target_name}", then "{input_name}" must be required as well'
                f' ("{input_name}: Required[<Type>]"). Alternatively, "{target_name}" should be optional'
                f' ("{target_name}: NotRequired[<Type>]") or an "MappingHint.aliasOf" might be missing.'
            )


def _validation_type_error(path: AttributePath, message: str) -> None:
    raise ManifestParseException(
        f'The attribute "{path.path}" did not have a valid structure/type: {message}'
    )


def _is_two_arg_x_list_x(t_args: Tuple[Any, ...]) -> bool:
    if len(t_args) != 2:
        return False
    lhs, rhs = t_args
    if get_origin(lhs) == list:
        if get_origin(rhs) == list:
            # It could still match X, List[X] - but we do not allow this case for now as the caller
            # does not support it.
            return False
        l_args = get_args(lhs)
        return bool(l_args and l_args[0] == rhs)
    if get_origin(rhs) == list:
        r_args = get_args(rhs)
        return bool(r_args and r_args[0] == lhs)
    return False


def _extract_typed_dict(
    base_type,
    default_target_attribute: Optional[str],
) -> Tuple[Optional[Type[TypedDict]], Any]:
    if is_typeddict(base_type):
        return base_type, None
    _, origin, args = unpack_type(base_type, False)
    if origin != Union:
        if isinstance(base_type, type) and issubclass(base_type, (dict, Mapping)):
            raise ValueError(
                "The source_format cannot be nor contain a (non-TypedDict) dict"
            )
        return None, base_type
    typed_dicts = [x for x in args if is_typeddict(x)]
    if len(typed_dicts) > 1:
        raise ValueError(
            "When source_format is a Union, it must contain at most one TypedDict"
        )
    typed_dict = typed_dicts[0] if typed_dicts else None

    if any(x is None or x is _NONE_TYPE for x in args):
        raise ValueError(
            "The source_format cannot be nor contain Optional[X] or Union[X, None]"
        )

    if any(
        isinstance(x, type) and issubclass(x, (dict, Mapping))
        for x in args
        if x is not typed_dict
    ):
        raise ValueError(
            "The source_format cannot be nor contain a (non-TypedDict) dict"
        )
    remaining = [x for x in args if x is not typed_dict]
    has_target_attribute = False
    anno = None
    if len(remaining) == 1:
        base_type, anno, _ = _parse_type(
            "source_format alternative form",
            remaining[0],
            forbid_optional=True,
            parsing_typed_dict_attribute=False,
        )
        has_target_attribute = bool(anno) and any(
            isinstance(x, TargetAttribute) for x in anno
        )
        target_type = base_type
    else:
        target_type = Union[tuple(remaining)]

    if default_target_attribute is None and not has_target_attribute:
        raise ValueError(
            'The alternative format must be Union[TypedDict,Annotated[X, DebputyParseHint.target_attribute("...")]]'
            " OR the parsed_content format must have exactly one attribute that is required."
        )
    if anno:
        final_anno = [target_type]
        final_anno.extend(anno)
        return typed_dict, Annotated[tuple(final_anno)]
    return typed_dict, target_type


def _dispatch_parse_generator(
    dispatch_type: Type[DebputyDispatchableType],
) -> Callable[[Any, AttributePath, Optional["ParserContextData"]], Any]:
    def _dispatch_parse(
        value: Any,
        attribute_path: AttributePath,
        parser_context: Optional["ParserContextData"],
    ):
        assert parser_context is not None
        dispatching_parser = parser_context.dispatch_parser_table_for(dispatch_type)
        return dispatching_parser.parse_input(
            value, attribute_path, parser_context=parser_context
        )

    return _dispatch_parse


def _dispatch_parser(
    dispatch_type: Type[DebputyDispatchableType],
) -> AttributeTypeHandler:
    return AttributeTypeHandler(
        dispatch_type.__name__,
        lambda *a: None,
        mapper=_dispatch_parse_generator(dispatch_type),
    )


def _parse_type(
    attribute: str,
    orig_td: Any,
    forbid_optional: bool = True,
    parsing_typed_dict_attribute: bool = True,
) -> Tuple[Any, Tuple[Any, ...], bool]:
    td, v, args = unpack_type(orig_td, parsing_typed_dict_attribute)
    md: Tuple[Any, ...] = tuple()
    optional = False
    if v is not None:
        if v == Annotated:
            anno = get_args(td)
            md = anno[1:]
            td, v, args = unpack_type(anno[0], parsing_typed_dict_attribute)

        if td is _NONE_TYPE:
            raise ValueError(
                f'The attribute "{attribute}" resolved to type "None".  "Nil" / "None" fields are not allowed in the'
                " debputy manifest, so this attribute does not make sense in its current form."
            )
        if forbid_optional and v == Union and any(a is _NONE_TYPE for a in args):
            raise ValueError(
                f'Detected use of Optional in "{attribute}", which is not allowed here.'
                " Please use NotRequired for optional fields"
            )

    return td, md, optional


def _normalize_attribute_name(attribute: str) -> str:
    if attribute.endswith("_"):
        attribute = attribute[:-1]
    return attribute.replace("_", "-")


@dataclasses.dataclass
class DetectedDebputyParseHint:
    target_attribute: str
    source_manifest_attribute: Optional[str]
    conflict_with_source_attributes: FrozenSet[str]
    conditional_required: Optional[ConditionalRequired]
    applicable_as_path_hint: bool

    @classmethod
    def parse_annotations(
        cls,
        anno: Tuple[Any, ...],
        error_context: str,
        default_attribute_name: Optional[str],
        is_required: bool,
        default_target_attribute: Optional[str] = None,
        allow_target_attribute_annotation: bool = False,
        allow_source_attribute_annotations: bool = False,
    ) -> "DetectedDebputyParseHint":
        target_attr_anno = find_annotation(anno, TargetAttribute)
        if target_attr_anno:
            if not allow_target_attribute_annotation:
                raise ValueError(
                    f"The DebputyParseHint.target_attribute annotation is not allowed in this context.{error_context}"
                )
            target_attribute = target_attr_anno.attribute
        elif default_target_attribute is not None:
            target_attribute = default_target_attribute
        elif default_attribute_name is not None:
            target_attribute = default_attribute_name
        else:
            if default_attribute_name is None:
                raise ValueError(
                    "allow_target_attribute_annotation must be True OR "
                    "default_attribute_name/default_target_attribute must be not None"
                )
            raise ValueError(
                f"Missing DebputyParseHint.target_attribute annotation.{error_context}"
            )
        source_attribute_anno = find_annotation(anno, ManifestAttribute)
        _source_attribute_allowed(
            allow_source_attribute_annotations, error_context, source_attribute_anno
        )
        if source_attribute_anno:
            source_attribute_name = source_attribute_anno.attribute
        elif default_attribute_name is not None:
            source_attribute_name = _normalize_attribute_name(default_attribute_name)
        else:
            source_attribute_name = None
        mutual_exclusive_with_anno = find_annotation(anno, ConflictWithSourceAttribute)
        if mutual_exclusive_with_anno:
            _source_attribute_allowed(
                allow_source_attribute_annotations,
                error_context,
                mutual_exclusive_with_anno,
            )
            conflicting_attributes = mutual_exclusive_with_anno.conflicting_attributes
        else:
            conflicting_attributes = frozenset()
        conditional_required = find_annotation(anno, ConditionalRequired)

        if conditional_required and is_required:
            if default_attribute_name is None:
                raise ValueError(
                    f"is_required cannot be True without default_attribute_name being not None"
                )
            raise ValueError(
                f'The attribute "{default_attribute_name}" is Required while also being conditionally required.'
                ' Please make the attribute "NotRequired" or remove the conditional requirement.'
            )

        not_path_hint_anno = find_annotation(anno, NotPathHint)
        applicable_as_path_hint = not_path_hint_anno is None

        return DetectedDebputyParseHint(
            target_attribute=target_attribute,
            source_manifest_attribute=source_attribute_name,
            conflict_with_source_attributes=conflicting_attributes,
            conditional_required=conditional_required,
            applicable_as_path_hint=applicable_as_path_hint,
        )


def _source_attribute_allowed(
    source_attribute_allowed: bool,
    error_context: str,
    annotation: Optional[DebputyParseHint],
) -> None:
    if source_attribute_allowed or annotation is None:
        return
    raise ValueError(
        f'The annotation "{annotation}" cannot be used here. {error_context}'
    )



from typing import (
    List,
    TypedDict,
    NotRequired,
    Annotated,
    Union,
    Mapping,
)

import pytest

from debputy.highlevel_manifest import PackageTransformationDefinition
from debputy.manifest_parser.tagging_types import (
    DebputyParsedContent,
    TypeMapping,
)
from debputy.manifest_parser.parse_hints import DebputyParseHint
from debputy.manifest_parser.declarative_parser import ParserGenerator
from debputy.manifest_parser.mapper_code import type_mapper_str2package
from debputy.manifest_parser.parser_data import ParserContextData
from debputy.manifest_parser.util import AttributePath
from debputy.packages import BinaryPackage
from debputy.substitution import NULL_SUBSTITUTION
from tutil import faked_binary_package


class TFinalEntity(DebputyParsedContent):
    sources: List[str]
    install_as: NotRequired[str]
    into: NotRequired[List[BinaryPackage]]
    recursive: NotRequired[bool]


class TSourceEntity(TypedDict):
    sources: NotRequired[List[str]]
    source: NotRequired[Annotated[str, DebputyParseHint.target_attribute("sources")]]
    as_: NotRequired[
        Annotated[
            str,
            DebputyParseHint.target_attribute("install_as"),
            DebputyParseHint.conflicts_with_source_attributes("sources"),
        ]
    ]
    into: NotRequired[Union[BinaryPackage, List[BinaryPackage]]]
    recursive: NotRequired[bool]


TSourceEntityAltFormat = Union[TSourceEntity, List[str], str]


foo_package = faked_binary_package("foo")
context_packages = {
    foo_package.name: foo_package,
}
context_package_states = {
    p.name: PackageTransformationDefinition(
        p,
        NULL_SUBSTITUTION,
        False,
    )
    for p in context_packages.values()
}


class TestParserContextData(ParserContextData):
    @property
    def _package_states(self) -> Mapping[str, PackageTransformationDefinition]:
        return context_package_states

    @property
    def binary_packages(self) -> Mapping[str, BinaryPackage]:
        return context_packages


@pytest.fixture
def parser_context():
    return TestParserContextData()


@pytest.mark.parametrize(
    "source_payload,expected_data,expected_attribute_path,parse_content,source_content",
    [
        (
            {"sources": ["foo", "bar"]},
            {"sources": ["foo", "bar"]},
            {
                "sources": "sources",
            },
            TFinalEntity,
            None,
        ),
        (
            {"sources": ["foo", "bar"], "install-as": "as-value"},
            {"sources": ["foo", "bar"], "install_as": "as-value"},
            {"sources": "sources", "install_as": "install-as"},
            TFinalEntity,
            None,
        ),
        (
            {"sources": ["foo", "bar"], "install-as": "as-value", "into": ["foo"]},
            {
                "sources": ["foo", "bar"],
                "install_as": "as-value",
                "into": [foo_package],
            },
            {"sources": "sources", "install_as": "install-as", "into": "into"},
            TFinalEntity,
            None,
        ),
        (
            {"source": "foo", "as": "as-value", "into": ["foo"]},
            {
                "sources": ["foo"],
                "install_as": "as-value",
                "into": [foo_package],
            },
            {"sources": "source", "install_as": "as", "into": "into"},
            TFinalEntity,
            TSourceEntity,
        ),
        (
            {"source": "foo", "as": "as-value", "into": ["foo"]},
            {
                "sources": ["foo"],
                "install_as": "as-value",
                "into": [foo_package],
            },
            {"sources": "source", "install_as": "as", "into": "into"},
            TFinalEntity,
            TSourceEntityAltFormat,
        ),
        (
            ["foo", "bar"],
            {
                "sources": ["foo", "bar"],
            },
            {"sources": "parse-root"},
            TFinalEntity,
            TSourceEntityAltFormat,
        ),
        (
            "foo",
            {
                "sources": ["foo"],
            },
            {"sources": "parse-root"},
            TFinalEntity,
            TSourceEntityAltFormat,
        ),
        (
            "foo",
            {
                "sources": ["foo"],
            },
            {"sources": "parse-root"},
            TFinalEntity,
            str,
        ),
        (
            ["foo", "bar"],
            {
                "sources": ["foo", "bar"],
            },
            {"sources": "parse-root"},
            TFinalEntity,
            List[str],
        ),
        (
            "foo",
            {
                "sources": ["foo"],
            },
            {"sources": "parse-root"},
            TFinalEntity,
            Union[str, List[str]],
        ),
        (
            ["foo", "bar"],
            {
                "sources": ["foo", "bar"],
            },
            {"sources": "parse-root"},
            TFinalEntity,
            Union[str, List[str]],
        ),
        (
            {"source": "foo", "recursive": True},
            {
                "sources": ["foo"],
                "recursive": True,
            },
            {"sources": "source", "recursive": "recursive"},
            TFinalEntity,
            TSourceEntityAltFormat,
        ),
    ],
)
def test_declarative_parser_ok(
    attribute_path: AttributePath,
    parser_context: ParserContextData,
    source_payload,
    expected_data,
    expected_attribute_path,
    parse_content,
    source_content,
):
    pg = ParserGenerator()
    pg.register_mapped_type(TypeMapping(BinaryPackage, str, type_mapper_str2package))
    parser = pg.generate_parser(
        parse_content,
        source_content=source_content,
    )
    data_path = attribute_path["parse-root"]
    parsed_data = parser.parse_input(
        source_payload, data_path, parser_context=parser_context
    )
    assert expected_data == parsed_data
    attributes = {k: data_path[k].name for k in expected_attribute_path}
    assert attributes == expected_attribute_path
